//
//  BeadType.swift
//  Oritatami Builder
//
//  Created by Nikao on 09/01/2019.
//  Copyright © 2019 Nikao. All rights reserved.
//

import Foundation
import UIKit

struct BeadType: CustomStringConvertible, Equatable {
    let id: Int
    let name: String
    let fixedLengthName: String
    let categoryId: Int
    let number: Int
    var color : __Color { didSet { updateColoredFixedLengthName() } }
    var description : String {
        return "\(name)[\(id)]<\(categoryId):\(number):\(coloredFixedLengthName)>"
    }
    var coloredFixedLengthName : String = ""
    
    init(id i: Int,
         name n: String,
         categoryId ct: Int, number nb: Int, color c: __Color) {
        self.id = i
        self.name = n
        self.categoryId = ct
        self.number = nb
        self.color = c
        
        let shortNameLength = Defaults.nameFixedLength
        let nameLength = name.count
        if nameLength > shortNameLength {
            self.fixedLengthName = name.prefix(shortNameLength-1) + "…"
        } else {
            let k = (shortNameLength - nameLength) /  2
            self.fixedLengthName = String(repeating: " ", count: k) + name + String(repeating: " ", count: shortNameLength - nameLength - k)
        }
        updateColoredFixedLengthName()
    }
    
    static func == (lhs: BeadType, rhs: BeadType) -> Bool { return lhs.id == rhs.id }
    
    mutating func updateColoredFixedLengthName() {
        coloredFixedLengthName = ANSITermCode.backgroundColor216(color) + fixedLengthName
    }
    
}

struct BeadTypeCategory: CustomStringConvertible {
    let id: Int
    let name: String
    var description : String {
        return "\(name)[\(id)]"
    }
    var color : __Color
    
    init(id i: Int, name n: String, color c: __Color? = nil) {
        self.id = i
        self.name = n
        
        self.color = c ?? BeadTypeCategory.defaultColor(categoryId: i)
    }
    
    static func defaultColor(categoryId: Int) -> __Color {
        let φ : Double = (1+sqrt(5))/2
        let hue = CGFloat(fmod(Double(categoryId) * φ, 1))
        return __Color(hue: hue, saturation: 1, brightness: 1, alpha: 1)
    }
}


class OritatamiRule: CustomStringConvertible, Codable {
    static let maxSize = BeadTypeSet.Defaults.maxSize
    let size: Int = OritatamiRule.maxSize
    var rule: [Bool] = Array(repeating: false, count: OritatamiRule.maxSize * OritatamiRule.maxSize)
    
    required init() {
    }
    
    @inline(__always) func index(_ i:Int, _ j: Int) -> Int {
        return i * size + j
    }
    
    @inline(__always) func attraction(between a: BeadType, and b: BeadType) -> Bool {
        return attraction(between: a.id, and:b.id)
    }
    
    @inline(__always) func attraction(between i: Int, and j: Int) -> Bool {
        return rule[index(i,j)]
    }
    
    @inline(__always) func setAttraction(between i: Int, and j: Int, attract: Bool = true) {
        rule[index(i,j)] = attract
        rule[index(j,i)] = attract
    }
    
    var description: String {
        var s = ""
        for i in 0..<size {
            for j in i..<size {
                if attraction(between: i, and: j) {
                    s += "\t\(i)❤️\(j)"
                }
            }
        }
        return s
    }
    
    func prettyDescription(with beadTypeSet: BeadTypeSet) -> String {
        let dict = encodeAsdictionnary(with: beadTypeSet)
        let sortedKeys = dict.keys.sorted(by: { $0.compare($1, options: .numeric).rawValue < 0 } ) // .sorted()
        var s = ""
        for k in sortedKeys {
            let attracted = dict[k]!.sorted(by: { $0.compare($1, options: .numeric).rawValue < 0 } )
            s += k + "❤️" + (attracted.reduce("") { $0+","+$1 }).dropFirst() + "\n"
        }
        return s
    }
    
    func encodeAsdictionnary(with beadTypeSet: BeadTypeSet) -> [String:[String]] {
        var dict : [String : [String]] = [:]
        for i in 0..<size {
            var attracted = [String]()
            for j in 0..<size {
                if attraction(between: i, and: j) {
                    attracted.append(beadTypeSet.beadType[j].name)
                }
            }
            if attracted.count > 0 {
                
                let c = beadTypeSet.beadType[i].color
                var (r,g,b) : (CGFloat,CGFloat,CGFloat) = (0,0,0)
                c.getRed(&r, green: &g, blue: &b, alpha: nil)
                let (rr,gg,bb) = (Int(255*r),Int(255*g),Int(255*b))
                let latexColor = "" // "\n\\definecolor{color" + beadTypeSet.beadType[i].name + "}{RGB}{" + rr.description + "," + gg.description + "," + bb.description + "}\n"
                let name = beadTypeSet.beadType[i].name + latexColor
                dict[name] = attracted
            }
        }
        return dict
    }
    
    func encodeAsListOfPairs(with beadTypeSet: BeadTypeSet) -> [[String]] {
        var pairs : [[String]] = []
        for i in 0..<size {
            for j in i..<size {
                if attraction(between: i, and: j) {
                    pairs.append([beadTypeSet.beadType[i].name, beadTypeSet.beadType[j].name])
                }
            }
        }
        return pairs
    }
    
    func encode(to encoder: Encoder) throws {
        guard let beadTypeSet = encoder.userInfo[beadTypeSetUserInfoKey] as? BeadTypeSet else { throw CodingErrors.encodingError("/!\\ Encoding failed: Bead Type Set missing")}
        let dict = encodeAsdictionnary(with: beadTypeSet)
        try dict.encode(to: encoder)
    }
    
    convenience init(from dict: [String:[String]], with beadTypeSet: BeadTypeSet) {
        self.init()
        for name in dict.keys {
            let b1 = beadTypeSet.beadType(parsing: name)
            for other in dict[name]! {
                let b2 = beadTypeSet.beadType(parsing: other)
                self.setAttraction(between: b1.id, and: b2.id, attract: true)
            }
        }
    }
    
    required convenience init(from decoder: Decoder) throws {
        let beadTypeSet = (decoder.userInfo[beadTypeSetUserInfoKey] as? BeadTypeSet) ?? BeadTypeSet()
        if let dict = try? Dictionary<String,[String]>(from: decoder) {
            self.init(from: dict, with: beadTypeSet)
        } else {
            throw CodingErrors.decodingError("No dictionnary")
        }
    }
}

class BeadTypeSet : CustomStringConvertible {
    
    var category: [BeadTypeCategory] = []
    var categoryIdFromName: [String:BeadTypeId] = [:]
    var beadTypeIdsInCategory: [[Int:BeadTypeId]] = []
    var categoryCount : Int { return category.count }
    
    var beadType : [BeadType] = []
    var beadTypeIdFromName: [String:BeadTypeId] = [:]
    var count : Int { return beadType.count }
    
    init() {
    }
    
    func category(named name: String) -> Int {
        // Create the category if necessary
        
        if let c = categoryIdFromName[name] {
            return c
        }
        
        let id = categoryCount
        let cat = BeadTypeCategory(id: id, name: name)
        category.append(cat)
        categoryIdFromName[name] = id
        
        beadTypeIdsInCategory.append([:])
        
        return id
    }
    
    func setColor(_ color: __Color, toCategoryNamed name: String) {
        let id = category(named: name)
        category[id].color = color
        for (number, b) in beadTypeIdsInCategory[id] {
            beadType[b].color = defaultColor(categoryId: id, number: number)
        }
    }
    
    @inline(__always) func beadTypeName(categoryId: Int, number: Int) -> String? {
        guard categoryId < categoryCount else { return nil }
        return "\(category[categoryId].name)\(number)"
    }
    
    @inline(__always) func beadTypeTwoLinesName(categoryId: Int, number: Int) -> (String,String)? {
        guard categoryId < categoryCount else { return nil }
        return (category[categoryId].name,"\(number)")
    }
    
    @inline(__always) func defaultColor(categoryId: Int, number: Int) -> __Color {
        let φ : Double = (1+sqrt(5))/2
        var hue : CGFloat = 0
        category[categoryId].color.getHue(&hue, saturation: nil, brightness: nil, alpha: nil)
        let saturation = CGFloat(0.2 + 0.8*fmod(Double(number) * φ, 1))
        return __Color(hue: hue, saturation: saturation, brightness: 1, alpha: 1)
        //        let c = __Color(hue:hue, saturation: 1, brightness: 1, alpha: 1)
        //        var r = CGFloat(0)
        //        var g = CGFloat(0)
        //        var b = CGFloat(0)
        //        var a = CGFloat(0)
        //        c.getRed(&r, green: &g, blue: &b, alpha: &a)
        //        let lum = sqrt( 0.299*r*r + 0.587*g*g + 0.114*b*b )
        //        let x = (1-0.25*CGFloat(fmod(Double(number) * phi, 1)))/lum
        //        print(r*x, g*x, b*x)
        //        return __Color(red: r*x, green: g*x, blue: b*x, alpha: 1)
    }
    
    @inline(__always) func beadType(named name: String) -> BeadType? {
        guard let id = beadTypeIdFromName[name] else { return nil }
        return beadType[id]
    }
    
    func beadType(category categoryName: String, number: Int) -> BeadType {
        // create if necessary
        let c = category(named: categoryName)
        
        if let id = beadTypeIdsInCategory[c][number] {
            // existe déjà
            return beadType[id]
        }
        
        let id = count
        let name = beadTypeName(categoryId: c, number: number)!
        if let other = beadTypeIdFromName[name] {
            let b = beadType[other]
            print("/!\\ Conflicting bead type names between \(b) and \((id,category[c].name,number,name))")
        }
        
        let color = defaultColor(categoryId: c, number: number)
        let b = BeadType(id: id, name: name,
                         categoryId: c, number: number, color: color)
        beadType.append(b)
        beadTypeIdsInCategory[c][number] = id
        beadTypeIdFromName[name] = id
        
        return b
    }
    
    func beadType(parsing name: String) -> BeadType {
        let (categoryName, number) = BeadTypeSet.parseBeadTypeName(name)
        return beadType(category: categoryName, number: number)
    }
    
    var description: String {
        return "Categories: \(category)\nBead types: \(beadType)"
    }
    
    static func parseBeadTypeName(_ name: String) -> (category: String, number: Int) {
        let string = name as NSString
        let regexpr = try! NSRegularExpression(pattern:"[0-9]+")
        let matches = regexpr.matches(in: name, options: [], range: NSRange(location:0, length: string.length))
        
        if matches.count == 0 {
            return (name, 0)
        }
        
        let numberRange = matches.last!.range
        let numberString = string.substring(with: numberRange)
        let number = Int(numberString)!
        let categoryName = string.substring(to: numberRange.location)
        
//        NSLog("\(categoryName, number)")
        
        return (categoryName, number)
    }
}

enum IdSequence {
    
    case id(Int), ids([Int]), value(String)
    indirect case array([IdSequence]), times(IdSequence, Int), declare(String, IdSequence), prefixOfInfiniteLoopOn(IdSequence, Int)
    
    private func expand(sequence: inout [Int], dictionary: inout [String: [Int]]) {
        switch self {
        case let .id(x): sequence.append(x)
        case let .ids(array): sequence.append(contentsOf: array)
        case let .value(name):
            if let seq = dictionary[name] {
                sequence.append(contentsOf: seq)
            }
        case let .array(array):
            for s in array {
                s.expand(sequence: &sequence, dictionary: &dictionary)
            }
        case let .times(s,n):
            for _ in 0..<n {
                s.expand(sequence: &sequence, dictionary: &dictionary)
            }
        case let .prefixOfInfiniteLoopOn(s,n):
            let seq = s.sequence
            for i in 0..<n {
                sequence.append(seq[i%seq.count])
            }
            
        case let .declare(name, seq):
            var seqName = [Int]()
            seq.expand(sequence: &seqName, dictionary: &dictionary)
            dictionary[name] = seqName
        }
    }
    
    var sequence : [Int] {
        var seq = [Int]()
        var dict = [String:[Int]]()
        self.expand(sequence: &seq, dictionary: &dict)
        return seq
    }
    
}



class OritatamiSystem : Codable {
    
    var name: String = Defaults.name
    
    var rule: OritatamiRule
    var beadTypeSet: BeadTypeSet
    
    var delay: Int
    var beadTypeIdsSequence: [BeadTypeId]
    var seedPath: [Direction]
    var transcriptPrefixLength: Int = 0
    
    var beadType : [BeadType] { return beadTypeSet.beadType }
    var beadTypeIdFromName : [String:BeadTypeId] { return beadTypeSet.beadTypeIdFromName }
    
    var periodCount: Int = Defaults.periodCount
    var transcriptLength: Int { return (beadTypeIdsSequence.count - seedPath.count - 1) }
    var transcriptPeriodLength: Int { return (beadTypeIdsSequence.count - transcriptPrefixLength - seedPath.count - 1) }
    var beadSequenceCount: Int { return (periodCount > 0 ? beadTypeIdsSequence.count + (periodCount-1) * transcriptPeriodLength : Int.max)}
    
    init() {
        self.rule = OritatamiRule()
        self.beadTypeSet = BeadTypeSet()
        self.delay = 1
        self.beadTypeIdsSequence = []
        self.seedPath = []
    }
    
    @inline(__always) func beadTypeId(indexed j: Int) -> BeadTypeId {
        if j < beadTypeIdsSequence.count {
            return beadTypeIdsSequence[j]
        } else {
            let prefix = seedPath.count + transcriptPrefixLength
            let k = prefix + 1 + (j - (prefix + 1))%transcriptPeriodLength
            return beadTypeIdsSequence[k]
        }
    }
    
    func beadTypeId(named name: String) -> BeadTypeId {
        if let id = beadTypeIdFromName[name] {
            return id
        }
        let id = beadTypeSet.beadType(parsing: name).id
        return id
    }
    
    func setAttraction(between beadTypeName1: String, and beadTypeName2: String, attract: Bool) {
        let id1 = beadTypeId(named: beadTypeName1)
        let id2 = beadTypeId(named: beadTypeName2)
        rule.setAttraction(between: id1, and: id2, attract: attract)
    }
    
    func setAttraction(between id1: Int, and id2: Int, attract: Bool = true) {
        rule.setAttraction(between: id1, and: id2, attract: attract)
    }
    
    func setBeadTypeSequence(to beadTypeNamesSequence: [String]) {
        var seq = [BeadTypeId]()
        for name in beadTypeNamesSequence {
            seq.append(beadTypeId(named: name))
        }
        beadTypeIdsSequence = seq
    }
    
    func getBeads(categoryName: String, closedRange: ClosedRange<Int>) -> [BeadTypeId] {
        var ids = [BeadTypeId]()
        for i in closedRange {
            ids.append(beadTypeSet.beadType(parsing: "\(categoryName)\(i)").id)
        }
        return ids
    }
    
    func getBeads(categoryName: String, range: Range<Int>) -> [BeadTypeId] {
        var ids = [BeadTypeId]()
        for i in range {
            ids.append(beadTypeSet.beadType(parsing: "\(categoryName)\(i)").id)
        }
        return ids
    }
    
    func addAttraction(between ids1: [BeadTypeId], and ids2: [BeadTypeId], according: [(Int,Int)], shiftedBy: (Int, Int) = (0,0)) {
        let (s1,s2) = shiftedBy
        for (i,j) in according {
            rule.setAttraction(between: ids1[(i+s1)%ids1.count], and: ids2[(j+s2)%ids2.count], attract: true)
        }
        
        //        if let a = beadTypeIdFromName["U5"], let b = beadTypeIdFromName["M0"], rule.attraction(between: a, and: b) {
        //            print("STOP: \(ids1.map() { beadType[$0].name }), \(ids2.map() { beadType[$0].name }), \(according), \(shiftedBy) ")
        //            exit(0)
        //        }
    }
    
    func addAttractions(glider2: [BeadTypeId]) {
        for i in 0..<glider2.count {
            rule.setAttraction(between: glider2[i], and: glider2[(i+2)%glider2.count], attract: true)
        }
    }
    
    
    // Encode/Decode
    enum CodingKeys: String, CodingKey {
        case delay, rule, seedConformation, transcriptPrefix, transcriptPeriod, periodCount, name, categoryColors, compactTranscriptPrefix, compactTranscriptPeriod
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        
        // Name
        try container.encode(name, forKey: .name)
        
        // Delay
        try container.encode(delay, forKey: .delay)
        
        // Category colors
        var categoryColors : [String: ColorEncoding] = [:]
        for c in beadTypeSet.category {
            categoryColors[c.name] = .hsba(c.color.hsba())
        }
        try container.encode(categoryColors, forKey: .categoryColors)
        
        // Rule
        let ruleListOfPairs = rule.encodeAsListOfPairs(with: beadTypeSet)
        try container.encode(ruleListOfPairs, forKey: .rule)
        
        // SeedConformation
        let seed0 = beadTypeSet.beadType[beadTypeIdsSequence[0]].name
        var seedConformationArray : [String] = [ seed0 ]
        for i in 0..<seedPath.count {
            seedConformationArray.append(seedPath[i].description)
            seedConformationArray.append(beadTypeSet.beadType[beadTypeIdsSequence[i+1]].name)
        }
        try container.encode(seedConformationArray, forKey: .seedConformation)
        
        // transcript prefix
        let transcriptPrefix : [String] = Array(beadTypeIdsSequence[seedPath.count..<seedPath.count + transcriptPrefixLength]).map { beadTypeSet.beadType[$0].name }
        try container.encode(transcriptPrefix, forKey: .transcriptPrefix)
        
        // Transcript period
        var transcriptPeriod : [String] = []
        for i in seedPath.count+1+transcriptPrefixLength..<beadTypeIdsSequence.count {
            transcriptPeriod.append(beadTypeSet.beadType[beadTypeIdsSequence[i]].name)
        }
        try container.encode(transcriptPeriod, forKey: .transcriptPeriod)
        
        // Period count
        try container.encode(periodCount, forKey: .periodCount)
    }
    
    required convenience init(from decoder: Decoder) throws {
        self.init()
        
        let container = try decoder.container(keyedBy: CodingKeys.self)
        
        // Name
        if let n = try? container.decode(String.self, forKey: .name) {
            self.name = n
        } else {
            print("/!\\ No name found")
        }
        
        // Delay
        if let d = try? container.decode(Int.self, forKey: .delay) {
            self.delay = d
        } else {
            print("/!\\ No delay found")
        }
        
        // Category colors
        if let d = try? container.decode([String : ColorEncoding].self, forKey: .categoryColors) {
            for (name, value) in d {
                beadTypeSet.setColor(value.color, toCategoryNamed: name)
            }
        } else {
            print("Using default colors")
        }
        
        // Rule
        if let listOfPairs = try? container.decode([[String]].self, forKey: .rule) {
            for pair in listOfPairs {
                let b1 = beadTypeSet.beadType(parsing: pair[0])
                let b2 = beadTypeSet.beadType(parsing: pair[1])
//                NSLog("\(pair) \(b1.id, b2.id)")
                rule.setAttraction(between: b1.id, and: b2.id)
            }
        } else {
            print("/!\\ No rule found")
        }
        
        print(rule.prettyDescription(with: beadTypeSet))
        
        // Transcript
        var transcript : [Int] = []
        var marksPrefix = [(Int, String)]()
        var marksPeriod = [(Int, String)]()
        transcriptPrefixLength = 0

        // SeedConformation
        
        if let seedConformationArray = try? container.decode([String].self, forKey: .seedConformation) {
            
            let n = seedConformationArray.count/2
            var path = [Direction]()
            path.reserveCapacity(n)
            var result = [BeadTypeId]()
            result.reserveCapacity(n+1)
            result.append(beadTypeSet.beadType(parsing: seedConformationArray[0]).id)
            for i in 0..<seedConformationArray.count/2 {
                path.append(Direction(named: seedConformationArray[2*i+1]))
                result.append(beadTypeSet.beadType(parsing: seedConformationArray[2*i+2]).id)
            }
            transcript += result
            seedPath = path
        } else {
            print("/!\\ No seed conformation found")
        }
        
        // transcript prefix
        
        if let compactTranscriptPrefix = try? container.decode(CompactTranscript.self, forKey: .compactTranscriptPrefix) {
            let (transcriptPrefix, _marksPrefix) = compactTranscriptPrefix.expand()
            marksPrefix = _marksPrefix
            var result = [BeadTypeId]()
            result.reserveCapacity(transcriptPrefix.count)
            for name in transcriptPrefix {
                result.append(beadTypeSet.beadType(parsing: name).id)
            }
            transcript.append(contentsOf: result)
            transcriptPrefixLength = transcriptPrefix.count
        } else if let transcriptPrefix = try? container.decode([String].self, forKey: .transcriptPrefix) {
            for name in transcriptPrefix {
                transcript.append(beadTypeSet.beadType(parsing: name).id)
            }
            transcriptPrefixLength = transcriptPrefix.count
        } else {
            print("/!\\ No transcript prefix found")
        }

        
        // transcript period
        
        if let compactTranscriptPeriod = try? container.decode(CompactTranscript.self, forKey: .compactTranscriptPeriod) {
            let (transcriptPeriod, _marksPeriod) = compactTranscriptPeriod.expand()
            marksPeriod = _marksPeriod
            var result = [BeadTypeId]()
            result.reserveCapacity(transcriptPeriod.count)
            for name in transcriptPeriod {
                result.append(beadTypeSet.beadType(parsing: name).id)
            }
            transcript.append(contentsOf: result)
        } else if let transcriptPeriod = try? container.decode([String].self, forKey: .transcriptPeriod) {
            for name in transcriptPeriod {
                transcript.append(beadTypeSet.beadType(parsing: name).id)
            }
        } else {
            print("/!\\ No transcript period found")
        }
        
        // transcript period count
        if let transcriptPeriodCount = try? container.decode(Int.self, forKey: .periodCount) {
            periodCount = transcriptPeriodCount
        } else {
            print("/!\\ No transcript period count found")
        }

        beadTypeIdsSequence = transcript
        
        print("""
****
Marks prefix:
\(marksPrefix)
Marks period:
\(marksPeriod)
****
""")
        
//        var s = "Sequence = "
//        for i in 0..<beadSequenceCount {
//            s += beadTypeSet.beadType[beadTypeId(indexed: i)].name + ","
//        }
//        print(s)

    }
    
    
}

enum ColorEncoding : Codable {
    case name(String), rgba([CGFloat]), hsba([CGFloat])
    
    static let rgba = "rgba"
    static let hsba = "hsba"
    static let name = "name"
    
    enum CodingKeys: String, CodingKey {
        case rgba, name, hsba
    }
    
    var color: __Color {
        switch self {
        case let .name(name):
            let defaultColor = __Color.grey500
            if #available(iOS 11.0, *) {
                return __Color(named: name) ?? __Color.materialColorByName[name] ?? defaultColor
            } else {
                // Fallback on earlier versions
                return __Color.materialColorByName[name] ?? defaultColor
            }
        case let .rgba(rgba):
            return __Color(rgba: rgba)
        case let .hsba(hsba):
            return __Color(hsba: hsba)
        }
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case let .name(name):
            try container.encode(name, forKey: .name)
        case let .rgba(rgba):
            try container.encode(rgba, forKey: .rgba)
        case let .hsba(hsba):
            try container.encode(hsba, forKey: .hsba)
        }
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)

        if let name = try? container.decode(String.self, forKey: .name) {
            self = .name(name)
        } else if let hsba = try? container.decode([CGFloat].self, forKey: .hsba) {
            self = .hsba(hsba)
        } else if let rgba = try? container.decode([CGFloat].self, forKey: .rgba) {
            self = .rgba(rgba)
        } else {
            self = .name("")
        }
    }
}
