//
//  CompressedPath.swift
//  Simple OS Simulator
//
//  Created by Nikao on 30/05/2021.
//  Copyright © 2021 Nikao. All rights reserved.
//

import UIKit

enum CompressedDirection : CustomStringConvertible {
    case NW(Int), NE(Int), E(Int), SE(Int), SW(Int), W(Int), loop(Int, [CompressedDirection])
    
    init(_ d: Direction) {
        switch d {
        case .NW: self = .NW
        case .NE: self = .NE
        case .E: self = .E
        case .SE: self = .SE
        case .SW: self = .SW
        case .W: self = .W
        }
    }
    
    static let NW : CompressedDirection = .NW(1)
    static let NE : CompressedDirection = .NE(1)
    static let E : CompressedDirection = .E(1)
    static let SE : CompressedDirection = .SE(1)
    static let SW : CompressedDirection = .SW(1)
    static let W : CompressedDirection = .W(1)
    
    var description: String {
        switch self {
        case let .NW(n): return "NW" + (n > 1 ? "(\(n))" : "")
        case let .NE(n): return "NE" + (n > 1 ? "(\(n))" : "")
        case let .E(n): return "E" + (n > 1 ? "(\(n))" : "")
        case let .SE(n): return "SE" + (n > 1 ? "(\(n))" : "")
        case let .SW(n): return "SW" + (n > 1 ? "(\(n))" : "")
        case let .W(n): return "W" + (n > 1 ? "(\(n))" : "")
        case let .loop(n, a): return "Loop{\(n), \(a)}"
        }
    }

    var direction: Direction? {
        switch self {
        case let .NW(n): return n > 0 ? .NW : nil
        case let .NE(n): return n > 0 ? .NE : nil
        case let .E(n): return n > 0 ? .E : nil
        case let .SE(n): return n > 0 ? .SE : nil
        case let .SW(n): return n > 0 ? .SW : nil
        case let .W(n): return n > 0 ? .W : nil
        default: return nil
        }
    }

    var last_direction: Direction? {
        switch self {
        case let .NW(n): return n > 0 ? .NW : nil
        case let .NE(n): return n > 0 ? .NE : nil
        case let .E(n): return n > 0 ? .E : nil
        case let .SE(n): return n > 0 ? .SE : nil
        case let .SW(n): return n > 0 ? .SW : nil
        case let .W(n): return n > 0 ? .W : nil
        case let .loop(n, a):
            guard n > 0 && a.count > 0 else { return nil }
            return a[n%a.count].last_direction
        }
    }

    var increment : CompressedDirection {
        switch self {
        case let .NW(n): return .NW(n+1)
        case let .NE(n): return .NE(n+1)
        case let .E(n): return .E(n+1)
        case let .SE(n): return .SE(n+1)
        case let .SW(n): return .SW(n+1)
        case let .W(n): return .W(n+1)
        case let .loop(n, a): return .loop(n+1, a)
        }
    }
    
    var length: Int {
        switch self {
        case let .NW(n), let .NE(n), let .E(n), let .SE(n), let .SW(n), let .W(n): return n
        case let .loop(n, a):
            let al = a.map() { $0.length }
            let sum = al.reduce(0) { $0 + $1 }
            var s = sum * (n/a.count)
            for i in 0..<n%a.count {
                s += al[i]
            }
            return s
        }
    }
    
    var is_a_loop : Bool { switch self { case .loop(_, _): return true; default: return false } }
    
    func stroke(from p: IntPoint, in ctx: CGContext, lineTo: Bool = false) -> IntPoint {
        var p = p
        var q : IntPoint
        switch self {
        case let .NW(n), let .NE(n), let .E(n), let .SE(n), let .SW(n), let .W(n):
            q = p + n * self.direction!.intPoint
            if lineTo {
                ctx.addLine(to: q.point)
            } else {
                ctx.addLines(between: [p.point, q.point])
                ctx.strokePath()
            }
        case let .loop(n, a):
            ctx.move(to: p.point)
            for i in 0..<n {
                p = a[i%a.count].stroke(from: p, in: ctx, lineTo: true)
            }
            ctx.strokePath()
            q = p
        }
        return q
    }

    func end_and_bounds(from p: IntPoint) -> (IntPoint, CGRect) {
        var p = p
        switch self {
        case let .NW(n), let .NE(n), let .E(n), let .SE(n), let .SW(n), let .W(n):
            let q = p + n * self.direction!.intPoint
            let x_min = min(q.point.x, p.point.x)
            let y_min = min(q.point.y, p.point.y)
            let x_max = max(q.point.x, p.point.x)
            let y_max = max(q.point.y, p.point.y)
            return (q, CGRect(x: x_min, y: y_min, width: x_max - x_min, height: y_max - y_min))
        case let .loop(n, a):
            var bounds = CGRect(x: p.point.x, y: p.point.y, width: 0, height: 0)
            for i in 0..<n {
                let (q, rect) = a[i%a.count].end_and_bounds(from: p)
                bounds = bounds.union(rect)
                p = q
            }
            return (p, bounds)
        }
    }

    func bounds(from p: IntPoint) -> CGRect {
        let (_, rect) = end_and_bounds(from: p)
        return rect
    }
}

class CompressedPath : CustomStringConvertible {
    
    var path: [(CompressedDirection, UIColor)] = []
        
    func mix_color(length1 n: Int, color1: UIColor, color2: UIColor) -> UIColor {
        let u = 1/CGFloat(n+1)
        var h1 : CGFloat = 0
        var h2 : CGFloat = 0
        var s1 : CGFloat = 0
        var s2 : CGFloat = 0
        var b1 : CGFloat = 0
        var b2 : CGFloat = 0
        var a1 : CGFloat = 0
        var a2 : CGFloat = 0
        color1.getHue(&h1, saturation: &s1, brightness: &b1, alpha: &a1)
        color2.getHue(&h2, saturation: &s2, brightness: &b2, alpha: &a2)
        let h = h1*(1-u) + u*h2
        let b = b1*(1-u) + u*b2
        let s = s1*(1-u) + u*s2
        let a = a1*(1-u) + u*a2
        
        return UIColor(hue: h, saturation: s, brightness: b, alpha: a)
    }

    func push(_ d: Direction, color: UIColor) {
        // Test si la direction est la même ou bien si ça matche l'avant dernier pattern
        guard let (last, last_color) = path.last else {
            path.append((CompressedDirection(d), color))
            return
        }
        
        switch last {
        case let .loop(n, a):
            // it's a glider : a.count == 2
            if a[n%a.count].direction == d {
                // extend the glider
                let color2 = mix_color(length1: n, color1: last_color, color2: color)
                path[path.count-1] = (last.increment, color2)
                return
            }
        case let .NW(n), let .NE(n), let .E(n), let .SE(n), let .SW(n), let .W(n):
            // it's a direction
            if last.direction == d {
                // directions match
                let _color = mix_color(length1: n, color1: last_color, color2: color)
                path[path.count-1] = (last.increment, _color)
                return
            }
            if last.length == 1 && path.count >= 2 {
                // check the direction before
                let (last2, last_color2) = path[path.count-2]
                if last2.length == 1 && last2.direction == d {
                    // a new glider !
                    let _color = mix_color(length1: 1,
                                           color1: last_color,
                                           color2: last_color2)
                    let __color = mix_color(length1: 2,
                                            color1: _color,
                                            color2: color)
                    path.removeLast()
                    path.removeLast()
                    path.append((.loop(3, [last2, last]), __color))
                    return
                }
            }
        }
        // no compression detected: just add the direction
        path.append((CompressedDirection(d), color))
    }
    
    func stroke(from p: IntPoint, in ctx: CGContext) {
        var p = p
        for (d,c) in path {
            ctx.setStrokeColor(c.cgColor)
            p = d.stroke(from: p, in: ctx)
        }
    }
    
    func bounds(from p: IntPoint) -> CGRect {
        var rect = CGRect(x: p.point.x, y: p.point.y, width: 0, height: 0)
        var p = p
        for (d,_) in path {
            let (q, _rect) = d.end_and_bounds(from: p)
            rect = rect.union(_rect)
            p = q
        }
        return rect
    }
    
    var description: String {
        let output = path.map { $0.0 }
        return output.description
    }
}
