//
//  EncodingDecoding.swift
//  Oritatami Builder
//
//  Created by Nikao on 10/01/2019.
//  Copyright © 2019 Nikao. All rights reserved.
//

import Foundation
import UIKit

let beadTypeSetUserInfoKey = CodingUserInfoKey(rawValue: "beadTypeSetKey")!

enum CodingErrors: Error { case parsingError(String), encodingError(String), decodingError(String) }

extension BeadType: Codable {
    enum CodingKeys: String, CodingKey {
        case id, name, categoryId, number, hueSaturationBrightnessAlpha, fixedLengthName, coloredFixedLengthName
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(id, forKey: .id)
        try container.encode(name, forKey: .name)
        try container.encode(fixedLengthName, forKey: .fixedLengthName)
        try container.encode(coloredFixedLengthName, forKey: .coloredFixedLengthName)
        try container.encode(categoryId, forKey: .categoryId)
        try container.encode(number, forKey: .number)
        var h : CGFloat = 0
        var s : CGFloat = 0
        var b : CGFloat = 0
        var a : CGFloat = 0
        color.getHue(&h, saturation: &s, brightness: &b, alpha: &a)
        try container.encode([h,s,b,a], forKey: .hueSaturationBrightnessAlpha)
    }
    
    // ADD: transcriptPrefix, transcriptPeriod, numberOfPeriod (<= 0 for infinite)
    
    init(from decoder: Decoder) throws {
        let values = try decoder.container(keyedBy: CodingKeys.self)
        id = try values.decode(Int.self, forKey: .id)
        name = try values.decode(String.self, forKey: .name)
        fixedLengthName = try values.decode(String.self, forKey: .fixedLengthName)
        coloredFixedLengthName = try values.decode(String.self, forKey: .coloredFixedLengthName)
        categoryId = try values.decode(Int.self, forKey: .categoryId)
        number = try values.decode(Int.self, forKey: .number)
        let colorArray = try values.decode([CGFloat].self, forKey: .hueSaturationBrightnessAlpha)
        color = __Color(hue: colorArray[0], saturation: colorArray[1], brightness: colorArray[2], alpha: colorArray[3])
    }
}

extension __Color {
    func rgba() ->  [CGFloat] {
        var red : CGFloat = 0
        var green : CGFloat = 0
        var blue : CGFloat = 0
        var alpha : CGFloat = 0
        getRed(&red, green: &green, blue: &blue, alpha: &alpha)
        return [red, green, blue, alpha]
    }
    
    func hsba() ->  [CGFloat] {
        var hue : CGFloat = 0
        var saturation : CGFloat = 0
        var brightness : CGFloat = 0
        var alpha : CGFloat = 0
        getHue(&hue, saturation: &saturation, brightness: &brightness, alpha: &alpha)
        return [hue, saturation, brightness, alpha]
    }
    
    convenience init(rgba: [CGFloat]) {
        let red = rgba.count > 0 ? rgba[0] : 1
        let green = rgba.count > 1 ? rgba[1] : 1
        let blue = rgba.count > 2 ? rgba[2] : 1
        let alpha = rgba.count > 3 ? rgba[3] : 1

        self.init(red: red, green: green, blue: blue, alpha: alpha)
    }

    convenience init(hsba: [CGFloat]) {
        let hue = hsba.count > 0 ? hsba[0] : 1
        let saturation = hsba.count > 1 ? hsba[1] : 1
        let brightness = hsba.count > 2 ? hsba[2] : 1
        let alpha = hsba.count > 3 ? hsba[3] : 1

        self.init(hue: hue, saturation: saturation, brightness: brightness, alpha: alpha)
    }
}

extension BeadTypeCategory: Codable {
    enum CodingKeys: String, CodingKey {
        case id, name, color
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(id, forKey: .id)
        try container.encode(name, forKey: .name)
        let rgba = color.rgba()
        try container.encode(rgba, forKey: .color)
    }
    
    init(from decoder: Decoder) throws {
        let values = try decoder.container(keyedBy: CodingKeys.self)
        id = try values.decode(Int.self, forKey: .id)
        name = try values.decode(String.self, forKey: .name)
        if let rgba = try? values.decode([CGFloat].self, forKey: .color) {
            color = __Color(rgba: rgba)
        } else if let colorName = try? values.decode(String.self, forKey: .color), let c = __Color.materialColorByName[colorName] {
            color = c
        } else {
            color = BeadTypeCategory.defaultColor(categoryId: id)
        }
    }

}

extension IdSequence : Codable {
    
    enum CodingKeys: String, CodingKey {
        case id, ids, array, times, declare, value, sequence, prefix
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case let .id(x): try container.encode(x, forKey: CodingKeys.id)
        case let .ids(a): try container.encode(a, forKey: CodingKeys.ids)
        case let .array(a):
            var array = container.nestedUnkeyedContainer(forKey: CodingKeys.array)
            try array.encode(contentsOf: a)
        case let .times(s, n):
            try container.encode(n, forKey: CodingKeys.times)
            try container.encode(s, forKey: CodingKeys.sequence)
        case let .declare(v, s):
            try container.encode(v, forKey: CodingKeys.declare)
            try container.encode(s, forKey: CodingKeys.sequence)
        case let .prefixOfInfiniteLoopOn(s,n):
            try container.encode(n, forKey: CodingKeys.prefix)
            try container.encode(s, forKey: CodingKeys.sequence)
        case let .value(v):
            try container.encode(v, forKey: CodingKeys.value)
        }
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        
        if let x = try? container.decode(Int.self, forKey: CodingKeys.id) {
            self = .id(x)
        } else if let a = try? container.decode([Int].self, forKey: CodingKeys.ids) {
            self = .ids(a)
        } else if let a = try? container.decode([IdSequence].self, forKey: CodingKeys.array) {
            self = .array(a)
        } else if let n = try? container.decode(Int.self, forKey: CodingKeys.times) {
            let s = try container.decode(IdSequence.self, forKey: CodingKeys.sequence)
            self = .times(s,n)
        } else if let v = try? container.decode(String.self, forKey: CodingKeys.declare) {
            let s = try container.decode(IdSequence.self, forKey: CodingKeys.sequence)
            self = .declare(v, s)
        } else if let v = try? container.decode(String.self, forKey: CodingKeys.value) {
            self = .value(v)
        } else if let n = try? container.decode(Int.self, forKey: CodingKeys.prefix) {
            let s = try container.decode(IdSequence.self, forKey: CodingKeys.sequence)
            self = .prefixOfInfiniteLoopOn(s,n)
        } else {
            throw CodingErrors.parsingError("Cannot parse encoding as IdSequence")
        }
    }
    
}

extension OritatamiSystem {
    func jsonEncoding(withName name: String = "OS") -> String {
        var jsonEncoding = "{\n"
        
        // Name
        jsonEncoding += "\"name\": \"\(name)\",\n"
        
        // Delay
        jsonEncoding += "\"delay\": \(delay),\n"
        
        // periodCount
        jsonEncoding += "\"periodCount\": \(periodCount),\n"
        
        
        // Rule
        jsonEncoding += "\"rule\":["
        for i in 0..<beadTypeSet.count {
            let nameI = beadTypeSet.beadType[i].name
            for j in i..<beadTypeSet.count {
                if rule.attraction(between: i, and: j) {
                    let nameJ = beadTypeSet.beadType[j].name
                    jsonEncoding += " [\"\(nameI)\",\"\(nameJ)\"],"
                }
            }
        }
        jsonEncoding += "],\n"

        // SeedConformation
        jsonEncoding += "\"seedConformation\": ["
        let bead0name = beadType[beadTypeIdsSequence[0]].name
        jsonEncoding += "\"\(bead0name)\","
        for i in 0..<seedPath.count {
            let dName = seedPath[i].description
            let beadIname = beadType[beadTypeIdsSequence[i+1]].name
            jsonEncoding += "\"\(dName)\",\"\(beadIname)\","
        }
        jsonEncoding += "],\n"
        
        // transcriptPeriod
        jsonEncoding += "\"transcriptPeriod\":["
        for i in seedPath.count+1..<beadTypeIdsSequence.count {
            let beadIname = beadType[beadTypeIdsSequence[i]].name
            jsonEncoding += "\"\(beadIname)\","
        }
        jsonEncoding += "],\n"

        // closing
        jsonEncoding += "}\n"
                
        return jsonEncoding
    }
}
