//
//  EnvironmentMask.swift
//  CAOS Simulator
//
//  Created by Nikao on 14/04/2020.
//  Copyright © 2020 Nikao. All rights reserved.
//

import Foundation

enum BitVectorStorage : Hashable, Equatable, Comparable {
    case uInt64(UInt64), uInt64array([UInt64])
    
    var count : Int {
        switch self {
        case .uInt64(_): return 1
        case let .uInt64array(a): return a.count
        }
    }
    
    func copy() -> BitVectorStorage {
        switch self {
        case let .uInt64(x):
            return .uInt64(x)
        case let .uInt64array(a):
            return .uInt64array(Array(a))
        }
    }
    
    static func == (lhs: BitVectorStorage, rhs: BitVectorStorage) -> Bool {
        switch lhs {
        case let .uInt64(x):
            switch rhs {
            case let .uInt64(y): return x == y
            default: return false
            }
        case let .uInt64array(a):
            switch rhs {
            case let .uInt64array(b):
                guard a.count == b.count else { return false }
                for i in 0..<a.count {
                    if a[i] != b[i] { return false }
                }
                return true
            default:
                return false
            }
        }
    }
    
    func hash(into hasher: inout Hasher) {
        switch self {
        case let .uInt64(x):
            hasher.combine(x)
        case let .uInt64array(a):
            hasher.combine(a.count)
            for x in a {
                hasher.combine(x)
            }
        }
    }
    
    static func < (lhs: BitVectorStorage, rhs: BitVectorStorage) -> Bool {
        guard lhs.count == rhs.count else { return lhs.count < rhs.count }
        switch lhs {
        case let .uInt64(x):
            switch rhs {
            case let .uInt64(y): return x < y
            case let .uInt64array(b): return x < b[0]
            }
        case let .uInt64array(a):
            switch rhs {
            case let .uInt64array(b):
                for i in 0..<a.count {
                    if a[i] < b[i] { return true }
                    if a[i] > b[i] { return false }
                }
                return false // lhs == rhs
            case let .uInt64(y):
                return a[0] < y
            }
        }
    }

    static func > (lhs: BitVectorStorage, rhs: BitVectorStorage) -> Bool {
        return rhs < lhs
    }

    static func <= (lhs: BitVectorStorage, rhs: BitVectorStorage) -> Bool {
        guard lhs.count == rhs.count else { return lhs.count < rhs.count }
        switch lhs {
        case let .uInt64(x):
            switch rhs {
            case let .uInt64(y): return x <= y
            case let .uInt64array(b): return x <= b[0]
            }
        case let .uInt64array(a):
            switch rhs {
            case let .uInt64array(b):
                for i in 0..<a.count {
                    if a[i] < b[i] { return true }
                    if a[i] > b[i] { return false }
                }
                return true // lhs == rhs
            case let .uInt64(y):
                return a[0] <= y
            }
        }
    }

    static func >= (lhs: BitVectorStorage, rhs: BitVectorStorage) -> Bool {
        return rhs <= lhs
    }
    
}

struct BitVector : Equatable, Hashable, Comparable {
    var stored: BitVectorStorage

    static let zeroInt64 = BitVector(size: 64)
    
    var size : Int
    var count : Int { return stored.count }
    
    init(size s: Int) {
        size = s
        stored = s <= 64 ?
            .uInt64(0) :
            .uInt64array(Array(repeating: UInt64(0),
                          count: Int(ceil(Double(s)/64))))
    }
    
    init(uInt64 i: UInt64 = 0) {
        stored = .uInt64(i)
        size = 64
    }
    
    init(uInt64array a: [UInt64]) {
        stored = .uInt64array(a)
        size = 64 * a.count
    }
    
    func copy() -> BitVector {
        switch stored {
        case let .uInt64(x):
            var copy = BitVector(uInt64: x)
            copy.size = self.size
            return copy
        case let .uInt64array(a):
            var copy = BitVector(uInt64array: Array(a))
            copy.size = self.size
            return copy
        }
    }
    
    mutating func write(from bv: BitVector) {
        switch stored {
        case .uInt64(_):
            switch bv.stored {
            case .uInt64(_):
                stored = bv.stored
            case let .uInt64array(ba):
                stored = .uInt64(ba[0])
            }
        case var .uInt64array(a):
            switch bv.stored {
            case let .uInt64(x):
                a[0] = x
            case let .uInt64array(ba):
                let k = min(count, ba.count)
                for i in 0..<k {
                    a[i] = ba[i]
                }
            }
            stored = .uInt64array(a)
        }
        
    }
    
    func permuted(by p:[Int]) -> BitVector {
        var result = BitVector(size:self.size)
        
        let c = p.count
        for i in 0..<size {
            result.setValue(of: p[i % c] + c * (i / c), to: self.value(i)) // permutation appliquée par bloc de taille c
        }
        
        return result
    }

    
    static func == (lhs: BitVector, rhs: BitVector) -> Bool {
        return lhs.stored == rhs.stored
    }
    
    static func < (lhs: BitVector, rhs: BitVector) -> Bool {
        return lhs.stored < rhs.stored
    }
    
    static func <= (lhs: BitVector, rhs: BitVector) -> Bool {
        return lhs.stored <= rhs.stored
    }
    
    static func > (lhs: BitVector, rhs: BitVector) -> Bool {
        return lhs.stored > rhs.stored
    }

    static func >= (lhs: BitVector, rhs: BitVector) -> Bool {
        return lhs.stored >= rhs.stored
    }

    func hash(into hasher: inout Hasher) {
        hasher.combine(stored)
    }

    func value(_ i: Int) -> Bool {
        switch stored {
        case let .uInt64(x): return (x & (UInt64(1)<<i)) != 0
        case let .uInt64array(a):
            let j = i%64
            let k = i/64
            if k >= a.count { return false }
            return (a[k] & (1<<j)) != 0
        }
    }
    
    
    mutating func setValue(of i: Int, to: Bool) {
        switch stored {
        case let .uInt64(x):
            let b : UInt64 = (UInt64(1))<<i
            stored = .uInt64(to ? x | b : x & ~b)
        case let .uInt64array(a):
            let k = i/64
            guard k < a.count else {
                print("/!\\ Try Setting bit out of range: \(k)/\(a.count)")
                return
            }
            var c : [UInt64] = Array(a)
            let b = UInt64(1)<<(i%64)
            c[k] = to ? (a[k] | b) : (a[k] & ~b)
            stored = .uInt64array(c)
        }
    }
    
}


class EnvironmentMask : Equatable, Hashable, CustomStringConvertible {
    
    // .Mark: EnvironmentMask class methods
    static var nameWidth = Defaults.nameWidth
        
    static func == (lhs: EnvironmentMask, rhs: EnvironmentMask) -> Bool {
        guard lhs.radius == rhs.radius else { return false }
        return lhs.bitVector == rhs.bitVector
    }

    @inline(__always) static func count(radius r: Int) -> Int {
        return 1 + 3 * r * (r + 1)
    }
    
    @inline(__always) static func cellIndex(radius r: Int, _ i: Int, _ j: Int) -> Int {
        let c = count(radius: r)
        let k = (r * i + (r + 1) * j) %% c
//        print(k)
        return k
    }
    
    static var cellPointsDict: [Int : [(x: Int, y: Int, index: Int)]] = [:]
    static func cellPoints(radius r: Int) -> [(Int,Int, Int)] {
        if cellPointsDict[r] == nil {
            var cp = [(Int,Int,Int)]()
            for j in -r...r {
                for i in max(-r, -r+j)...min(r,r+j) {
                    cp.append((i, j, cellIndex(radius: r, i, j)))
                }
            }
            cellPointsDict[r] = cp
        }
        return cellPointsDict[r]!
    }
    
    static var cwRotatePermutationDict: [Int:[Int]] = [:]
    static func cwRotatePermutation(radius r: Int) -> [Int] {
        if cwRotatePermutationDict[r] == nil {
            let c = count(radius: r)
            let cp = cellPoints(radius: r)
            var cwRotatePermutation = Array(repeating:0, count: c)
            
            let eCW = Direction.E.cw.intPoint
            let swCW = Direction.SW.cw.intPoint
            
            let eCWindex = cellIndex(radius: r, eCW.x, eCW.y)
            let swCWindex = cellIndex(radius: r, swCW.x, swCW.y)

            for (i,j,k) in cp {
                let cw = (i * eCWindex + j * swCWindex) %% c
                cwRotatePermutation[k] = cw
            }
            
            cwRotatePermutationDict[r] = cwRotatePermutation
        }
        return cwRotatePermutationDict[r]!
    }

    static var flipPermutationDict: [Int:[Int]] = [:]
    static func flipPermutation(radius r: Int) -> [Int] {
        if flipPermutationDict[r] == nil {
            let c = count(radius: r)
            let cp = cellPoints(radius: r)
            var flipPermutation = Array(repeating: 0, count: c)

            for (i,j,k) in cp {
                let s = cellIndex(radius: r, j, i)
                flipPermutation[k] = s
            }
            flipPermutationDict[r] = flipPermutation
        }
        return flipPermutationDict[r]!
    }

    // .Mark: EnvironmentMask methods
    var delay : Int { radius - 1}
    var radius : Int = Defaults.delay + 1
    var count : Int { EnvironmentMask.count(radius: radius)  }
    
    var bitVector = BitVector.zeroInt64         // accessible + attractions entre les beadtypes accessibles et les new beads = un vecteur de taille accessible + delay * accessible (1 for each new bead)
    var flipped : Bool = false
    var clockwiseRotation : Int = 0
    
    init(radius r: Int) {
        radius = r
        bitVector = BitVector(size: count)
    }
    

    @inline(__always) func cellIndex(_ i: Int, _ j:Int) -> Int {
        return EnvironmentMask.cellIndex(radius: radius, i, j)
    }
        
    @inline(__always) func cellIndex(_ p: IntPoint) -> Int {
        return cellIndex(p.x, p.y)
    }
        
    
    func computeAccessibleFrom(occupied: BitVector) -> BitVector {
        let c = count
        var accessible = BitVector(size: c)
        // look for accessible from (0,0): BFS at distance delay+1
        var distance = Array(repeating: Int.max, count: self.count)
        let kZero = cellIndex(0,0)
        var queue = FastQueue<Int>()
        queue.push(kZero)
        distance[kZero] = 0
        accessible.setValue(of: kZero, to: true)
        while !(queue.empty) {
            // pop only empty site in the queue but at start
            let k = queue.pop()!
            // index
            let dk = distance[k]
            
            if dk > delay {
                // Done with the exploration at distance <= delay
                break
            }
                        
            for d in Direction.all {
                let v = (k + cellIndex(d.intPoint)) % c
                if occupied.value(v) { // occupied
                    accessible.setValue(of: v, to: true)
                } else { // empty
                    if distance[v] > dk + 1 {
                        // not yet visited and must be visited
                        distance[v] = dk + 1
                        queue.push(v)
                    }
                }
            }
        }
        
        return accessible
    }
    
    func computeMinBitVectorByIsometries(of bv: BitVector) -> (min: BitVector, flipped: Bool, clockwiseRotation: Int)  {
        var minBitVector = bv.copy()
        var bitCode = bv.copy()
        
        let r = self.radius
        var flipped = false
        var clockwiseRotation = 0
        let cwRotatePermutation = EnvironmentMask.cwRotatePermutation(radius: r)
        let flipPermutation = EnvironmentMask.flipPermutation(radius: r)

        for i in 0...1 {
            for cw in 0..<Direction.count {
                if bitCode < minBitVector {
                    minBitVector = bitCode.copy()
                    clockwiseRotation = cw
                    flipped = (i == 1)
                }
                bitCode = bitCode.permuted(by: cwRotatePermutation)
            }
            bitCode = bitCode.permuted(by: flipPermutation)
        }
        
        return (min: minBitVector, flipped: flipped, clockwiseRotation: clockwiseRotation)
    }
    
    init(at p: IntPoint, lastBeadIndex: Int, grid: TriangularGridProtocol, oritatamiSystem os: OritatamiSystem)
    {
        self.radius = os.delay + 1
        
        let c = self.count
        let r = self.radius
        let delay = self.delay
        let cellPoints = EnvironmentMask.cellPoints(radius: r)
        
        var occupied = BitVector(size: c)
        // record occupied
        for (i,j,k) in cellPoints {
            let q = p + IntPoint(x: i, y: j)
            let b = grid.beadTypeId(at: q)
            occupied.setValue(of: k, to: b != nil)
        }

        let accessible = computeAccessibleFrom(occupied: occupied)
        print("acces", accessible, accessible.size)
        
        var bitCode = BitVector(size: c + c * delay)
        // copie accessible
        bitCode.write(from: accessible)
        print("bitcode", bitCode, bitCode.size)
        
        // record the attraction between the accessible and the upcoming beads
        let upcomingBeadTypesId : [BeadTypeId] = Array(1...delay).map {
            os.beadTypeIdsSequence[lastBeadIndex+$0]
        }
        for (i,j,k) in cellPoints {
            let q = p + IntPoint(x: i, y: j)
            let b = grid.beadTypeId(at: q)
            if let b = b, accessible.value(k) {
                for x in 0..<delay {
                    if os.rule.attraction(between: upcomingBeadTypesId[x],
                                          and: b) {
                        bitCode.setValue(of: c + x * c + k, to: true)
                    }
                }
            }
        }

        // Rotate/Sym the accessible bitvector to find the minimum bitvector value
        (bitVector, flipped, clockwiseRotation) = computeMinBitVectorByIsometries(of: bitCode)
    }
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(bitVector)
    }
    
    func description(nameWidth: Int = EnvironmentMask.nameWidth) -> String {
        var s = "flip:\(flipped) then cw:\(clockwiseRotation)\n"
        let r = self.radius
        let c = self.count
        let delay = self.delay
        let pad = String(repeating: " ", count: nameWidth+1)
        let pad2 = String(repeating: " ", count: nameWidth/2+1)
        for j in -r...r {
            let jabs = abs(j)
            s += String(repeating: pad, count: jabs/2) + (jabs%2 == 0 ? "" : pad2)
            for i in max(-r, -r+j)...min(r,r+j) {
                let k = cellIndex(i,j)
                var name = ""
                if bitVector.value(k) {
                    var attract = UInt64(0)
                    for x in 0..<delay {
                        if bitVector.value(c + x*c + k) {
                            attract += 1<<x
                        }
                    }
                    name = String(format:"%X", attract)
                } else {
                    name = "."
                }
                name = name.description.centered(width: nameWidth)
                s += name + " "
            }
            s.removeLast()
            s += "\n"
        }
        return s
    }
    
    var description: String {
        return description()
    }
}

class Environment {
    
    
    
    func nameBeadTypesVectors(index k: Int) -> String {
        guard k >= 0 else { return "???" } // k = index in the grid
        let c = Defaults.alphabet.count
        let a = Defaults.alphabet[k % c]
        if k < c { return a }
        if k < 2*c { return a + "'" }
        if k < 3*c { return a + "''" }
        return a + "\(k / c)"
    }
    
    var description : String {
        return ""
//        let n = name(index: k).padding(toLength: 3, withPad: " ", startingAt: 0)
    }
    

}
