//
//  IntPoint.swift
//  Scary Pacman
//
//  Created by Nikao on 19/9/17.
//  Copyright © 2017 Algoseum. All rights reserved.
//

import Foundation
import QuartzCore

enum Directions: CustomStringConvertible {
    case nE(Int), nSE(Int), nSW(Int), nW(Int), nNW(Int), nNE(Int), Dirs([Directions])
    var description: String {
        switch self {
        case let .nE(x): return "E*\(x)"
        case let .nSE(x): return "SE*\(x)"
        case let .nSW(x): return "SW*\(x)"
        case let .nW(x): return "W*\(x)"
        case let .nNW(x): return "NW*\(x)"
        case let .nNE(x): return "NE*\(x)"
        case let .Dirs(s): return s.description
        }
    }
    
    var mirror_EW: Directions {
        switch self {
        case let .nE(x): return .nW(x)
        case let .nSE(x): return .nSW(x)
        case let .nSW(x): return .nSE(x)
        case let .nW(x): return .nE(x)
        case let .nNW(x): return .nNE(x)
        case let .nNE(x): return .nNW(x)
        case let .Dirs(s): return .Dirs(s.map { $0.mirror_EW})
        }
    }
    
    var opposite: Directions {
        switch self {
        case let .nE(x): return .nW(x)
        case let .nSE(x): return .nNW(x)
        case let .nSW(x): return .nNE(x)
        case let .nW(x): return .nE(x)
        case let .nNW(x): return .nSE(x)
        case let .nNE(x): return .nSW(x)
        case let .Dirs(s): return .Dirs(s.map { $0.opposite })
        }
    }
    
    var cw: Directions {
        switch self {
        case let .nE(x): return .nSE(x)
        case let .nSE(x): return .nSW(x)
        case let .nSW(x): return .nW(x)
        case let .nW(x): return .nNW(x)
        case let .nNW(x): return .nNE(x)
        case let .nNE(x): return .nE(x)
        case let .Dirs(s): return .Dirs(s.map { $0.cw })
        }
    }
    
    func cw(_ r: Int) -> Self {
        var result = self
        for _ in 0..<r {
            result = result.cw
        }
        return result
    }
    
    var ccw: Directions {
        switch self {
        case let .nE(x): return .nNE(x)
        case let .nSE(x): return .nE(x)
        case let .nSW(x): return .nSE(x)
        case let .nW(x): return .nSW(x)
        case let .nNW(x): return .nW(x)
        case let .nNE(x): return .nNW(x)
        case let .Dirs(s): return .Dirs(s.map { $0.ccw })
        }
    }
    
    func ccw(_ r: Int) -> Self {
        var result = self
        for _ in 0..<r {
            result = result.ccw
        }
        return result
    }
    
    func expand() -> [Direction] {
        var n = 0
        var d = Direction.E
        switch self {
        case let .Dirs(s):
            var result = [Direction]()
            for x in s {
                result.append(contentsOf: x.expand())
            }
            return result
        case let .nE(x): n = x; d = .E
        case let .nSE(x): n = x; d = .SE
        case let .nSW(x): n = x; d = .SW
        case let .nW(x): n = x; d = .W
        case let .nNW(x): n = x; d = .NW
        case let .nNE(x): n = x; d = .NE
        }
        if n == 0 {
            return []
        }
        if n < 0 {
            n = -n
            d = d.opposite
        }
        return [Direction](repeating: d, count: n)
    }
}

typealias LocalConformation = (path: Directions, beads: [BeadTypeId])
typealias Conformation = [LocalConformation]



extension Conformation {
    func cw(_ r: Int) -> Self {
        return self.map { (path, beads) in (path.cw(r), beads) }
    }
    
    func ccw(_ r: Int) -> Self {
        return self.map { (path, beads) in (path.ccw(r), beads) }
    }
    
    var mirror_EW : Self {
        return self.map { (path, beads) in (path.mirror_EW, beads) }
    }
    
    var opposite : Self {
        return self.map { (path, beads) in (path.opposite, beads) }
    }
    
    func pathAndBeads() -> (Path, [BeadTypeId]) {
        var path = Path()
        var beads = [BeadTypeId]()
        
        for (p, b) in self {
            path.append(contentsOf: p.expand())
            beads.append(contentsOf: b)
        }
        
        return (path, beads)
    }
}

typealias LocalDoubleConformation = (path: Directions, beads1: [BeadTypeId], beads2: [BeadTypeId])
typealias DoubleConformation = [LocalDoubleConformation]

extension DoubleConformation {
    var conformation1 : Conformation {
        return self.map { ($0.path, $0.beads1) }
    }

    var conformation2 : Conformation {
        return self.map { ($0.path, $0.beads2) }
    }
    
    func cw(_ r: Int) -> Self {
        return self.map { (p, b1, b2) in (p.cw(r), b1, b2) }
    }
    
    func ccw(_ r: Int) -> Self {
        return self.map { (p, b1, b2) in (p.ccw(r), b1, b2) }
    }
    
    var mirror_EW : Self {
        return self.map { (p, b1, b2) in (p.mirror_EW, b1, b2) }
    }
    
    var opposite : Self {
        return self.map { (p, b1, b2) in (p.opposite, b1, b2) }
    }
    
    func pathAndBeadSequences() -> (Path, [BeadTypeId], [BeadTypeId]) {
        var path = Path()
        var beads1 = [BeadTypeId]()
        var beads2 = [BeadTypeId]()

        for (p, b1, b2) in self {
            path.append(contentsOf: p.expand())
            beads1.append(contentsOf: b1)
            beads2.append(contentsOf: b2)
        }
        
        return (path, beads1, beads2)
    }
    
    func pathAndMixedBeads(beads2Indices: [Int] = []) -> (Path, [BeadTypeId]) {
        let (path, beads1, beads2) = pathAndBeadSequences()
        
        var beads = beads1

        for i in beads2Indices {
            beads[i] = beads2[i]
        }
        
        return (path, beads)
    }
}

enum Direction: Int, CustomStringConvertible {
    case E, SE, SW, W, NW, NE
    
    var description: String {
        switch self {
        case .E: return "E"
        case .SE: return "SE"
        case .SW: return "SW"
        case .W: return "W"
        case .NW: return "NW"
        case .NE: return "NE"
        }
    }
    
    static let count = 6
    static let halfCount = Direction.count/2
    
    static let all: [Direction] = [.E, .SE, .SW, .W, .NW, .NE, ]
    
    var cw: Direction {
        return Direction(rawValue: (rawValue+1) % Direction.count)!
    }
    
    var ccw: Direction {
        return Direction(rawValue: (rawValue+Direction.count-1) % Direction.count)!
    }
    
    func cw(_ n: Int) -> Direction {
        return Direction(rawValue: (rawValue + (n%Direction.count)) % Direction.count)!
    }
    
    func ccw(_ n: Int) -> Direction {
        return Direction(rawValue: (rawValue + Direction.count - (n%Direction.count)) % Direction.count)!
    }
    
    var opposite: Direction {
        return Direction(rawValue: (rawValue+Direction.halfCount) % Direction.count)!
    }
    
    static prefix func - (d: Direction) -> Direction { return d.opposite }
    
    var intPoint: IntPoint {
        switch self {
        case .NW: return IntPoint(x: -1, y: -1) // NW = -E-SW
        case .NE: return IntPoint(x: 0, y: -1) // NE = -SW
        case .E: return IntPoint(x: 1, y: 0) // E
        case .SE: return IntPoint(x: 1, y: 1) // SE = SW + E
        case .SW: return IntPoint(x: 0, y: 1) // SW
        case .W: return IntPoint(x: -1, y: 0) // W = -E
        }
    }
    
    var point: CGPoint { return intPoint.point }
    
    var char: String {
        switch self {
        case .E, .W: return "-"
        case .NE, .SW: return "/"
        case .NW, .SE: return "\\"
        }
    }

    init(named name: String) {
        switch name {
        case "E": self = .E
        case "SE": self = .SE
        case "SW": self = .SW
        case "W": self = .W
        case "NW": self = .NW
        case "NE": self = .NE
        default: self = .E
        }
    }
    
}


typealias Path = [Direction]

extension Path {
    func cw(_ n : Int = 1) -> Path { return self.map { $0.cw(n) } }
    func ccw(_ n : Int = 1) -> Path { return self.map { $0.ccw(n) } }
}

struct IntPoint: CustomStringConvertible, Equatable, Hashable {
    var x: Int // E
    var y: Int // SW
    
    static let sqrt32: CGFloat = sqrt(3)/2
    
    static let zero = IntPoint(x:0, y:0)
    
    var description: String {
        return "(\(x),\(y))"
    }
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(x)
        hasher.combine(y)
    }
    
    var point: CGPoint {
        let xf  = CGFloat(x)
        let yf = CGFloat(y)
        return CGPoint(x: xf - yf / 2, y: yf * IntPoint.sqrt32)
        //        return CGPoint(x: CGFloat(x) + (y%2 == 0 ? 0 : 0.5), y: CGFloat(y)*IntPoint.sqrt32)
    }
    
    init(x: Int, y: Int) {
        self.x = x
        self.y = y
    }
    
    init(closestTo p: CGPoint) {
        y = Int(round(p.y/IntPoint.sqrt32))
        x = Int(round(p.x + CGFloat(y)/2))
    }
    
    func neighbor(_ d: Direction, by steps: Int = 1) -> IntPoint {
        return self + steps * d.intPoint
    }
    
    func NE(_ n:Int = 1) -> IntPoint { return neighbor(.NE, by: n) }
    
    func E(_ n:Int = 1) -> IntPoint { return neighbor(.E, by: n) }
    
    func SE(_ n:Int = 1) -> IntPoint { return neighbor(.SE, by: n) }
    
    func NW(_ n:Int = 1) -> IntPoint { return neighbor(.NW, by: n) }
    
    func W(_ n:Int = 1) -> IntPoint { return neighbor(.W, by: n) }
    
    func SW(_ n:Int = 1) -> IntPoint { return neighbor(.SW, by: n) }
    
    func dist2(to p: IntPoint) -> CGFloat {
        return point.dist2(to: p.point)
    }
    
    static prefix func - (p: IntPoint) -> IntPoint { return IntPoint(x: -p.x, y: -p.y) }
    static prefix func + (p: IntPoint) -> IntPoint { return p }
    
    static func + (lhs: IntPoint, rhs: IntPoint) -> IntPoint { return IntPoint(x: lhs.x + rhs.x, y: lhs.y + rhs.y) }
    static func - (lhs: IntPoint, rhs: IntPoint) -> IntPoint { return IntPoint(x: lhs.x - rhs.x, y: lhs.y - rhs.y) }
    static func == (lhs: IntPoint, rhs: IntPoint) -> Bool { return lhs.x == rhs.x && lhs.y == rhs.y }
    static func != (lhs: IntPoint, rhs: IntPoint) -> Bool { return lhs.x != rhs.x || lhs.y != rhs.y }
    static func * (lhs: Int, rhs: IntPoint) -> IntPoint {
        return IntPoint(x: lhs * rhs.x, y: lhs * rhs.y)
    }
    static func * (lhs: IntPoint, rhs: Int) -> IntPoint {
        return IntPoint(x: rhs * lhs.x, y: rhs * lhs.y)
    }
    
    func endOf(path: [Direction], shortenBy k: Int = 0) -> IntPoint {
        var p = self
        for i in k..<path.count - k  {
            p = p + path[i].intPoint
        }
        return p
    }
}

struct IntSize: CustomStringConvertible {
    var width: Int
    var height: Int
    
    static let zero = IntSize(width: 0, height: 0)
    
    var description: String {
        return "\(width)x\(height)"
    }
}

struct IntRect: CustomStringConvertible {
    
    var origin = IntPoint.zero
    var size = IntSize.zero
    
    var x : Int {
        get { return origin.x }
        set(x) { origin.x = x }
    }
    var y : Int {
        get { return origin.y }
        set(y) { origin.y = y }
    }
    
    var width : Int {
        get { return size.width }
        set(w) { size.width = w }
    }
    var height : Int {
        get { return size.height }
        set(h) { size.height = h }
    }
    
    var description: String {
        return "\(origin)-\(size)"
    }
    
    func contains(_ p: IntPoint) -> Bool { // must be strictly inside so that its neighbors are inside too
        return p.x > x && p.y > y && p.x < x - 1 + width && p.y < y - 1 + height
    }
    
    // EXTENSION
    
    mutating func extend(toInclude p: IntPoint) {
        while true {
            //        fprintf(stderr, "Putting point (%d,%d) inside {(x:%d,y:%d),(w:%d,h:%d)}\n", p.x, p.y, newFrame.origin.x, newFrame.origin.y, newFrame.size.w, newFrame.size.h);
            if p.x <= origin.x {
                origin.x -= size.width
                size.width *= 2
                continue
            }
            if p.x >= origin.x - 1 + size.width {
                size.width *= 2
                continue
            }
            if p.y <= origin.y {
                origin.y -= size.height
                size.height *= 2
                continue
            }
            if p.y >= origin.y - 1 + size.height {
                size.height *= 2
                continue
            }
            // Stop p is strictly inside now
            break;
        }
        //    fprintf(stderr, "point (%d,%d) is inside {(x:%d,y:%d),(w:%d,h:%d)", p.x, p.y, newFrame.origin.x, newFrame.origin.y, newFrame.size.w, newFrame.size.h);
    }
    
}

extension CGSize {
    static func * (lhs: CGSize, rhs: CGFloat) -> CGSize { return CGSize(width: lhs.width * rhs,
                                                                        height: lhs.height * rhs) }
    
    static func * (lhs: CGFloat, rhs: CGSize) -> CGSize { return CGSize(width: lhs * rhs.width,
                                                                        height: lhs * rhs.height) }
    
    static func / (lhs: CGSize, rhs: CGFloat) -> CGSize { return CGSize(width: lhs.width / rhs,
                                                                        height: lhs.height / rhs) }
    
}

extension CGPoint {
    func moveScale(_ origin: CGPoint = CGPoint.zero,_ scale: CGFloat = 1) -> CGPoint {
        return origin + scale * self
    }
    
    func unMoveScale(_ origin: CGPoint = CGPoint.zero,_ scale: CGFloat = 1) -> CGPoint {
        return (self - origin) / scale
    }
    
    // artihmetic point
    static prefix func - (p: CGPoint) -> CGPoint { return CGPoint(x: -p.x, y: -p.y) }
    static prefix func + (p: CGPoint) -> CGPoint { return p }
    
    static func + (lhs: CGPoint, rhs: CGPoint) -> CGPoint { return CGPoint(x: lhs.x + rhs.x, y: lhs.y + rhs.y) }
    static func - (lhs: CGPoint, rhs: CGPoint) -> CGPoint { return CGPoint(x: lhs.x - rhs.x, y: lhs.y - rhs.y) }
    
    static func * (lhs: CGPoint, rhs: CGFloat) -> CGPoint { return CGPoint(x: lhs.x * rhs, y: lhs.y * rhs) }
    static func * (lhs: CGFloat, rhs: CGPoint) -> CGPoint { return CGPoint(x: lhs * rhs.x, y: lhs * rhs.y) }
    
    static func / (lhs: CGPoint, rhs: CGFloat) -> CGPoint { return CGPoint(x: lhs.x / rhs, y: lhs.y / rhs) }
    
    
    // Dist
    
    func dist2(to p: CGPoint) -> CGFloat {
        let dx = p.x - x
        let dy = p.y - y
        return dx * dx + dy * dy
    }
    
}

extension CGRect {
    
    static func + (lhs: CGPoint, rhs: CGRect) -> CGRect { return CGRect(origin: lhs + rhs.origin, size: rhs.size) }
    static func + (lhs: CGRect, rhs: CGPoint) -> CGRect { return CGRect(origin: lhs.origin + rhs, size: lhs.size) }
    
    static func - (lhs: CGRect, rhs: CGPoint) -> CGRect { return CGRect(origin: lhs.origin - rhs, size: lhs.size) }
    
    static func * (lhs: CGRect, rhs: CGFloat) -> CGRect { return CGRect(origin: lhs.origin * rhs, size: rhs * lhs.size) }
    static func * (lhs: CGFloat, rhs: CGRect) -> CGRect { return CGRect(origin: rhs.origin * lhs, size: lhs * rhs.size) }
    static func / (lhs: CGRect, rhs: CGFloat) -> CGRect {
        return CGRect(origin: lhs.origin / rhs,
                      size: lhs.size / rhs)
    }
    
    func moveScale(_ translation: CGPoint,_ scale: CGFloat) -> CGRect {
        return CGRect(origin: translation + scale * origin,
                      size: scale * size)
    }
    
    func unMoveScale(_ translation: CGPoint,_ scale: CGFloat) -> CGRect {
        return CGRect(origin: (origin - translation) / scale,
                      size: size / scale)
    }
    
    
    init(for p:IntPoint) {
        self.init()
        let r = Defaults.intPointRadius
        origin = p.point - CGPoint(x: r, y:r)
        size = CGSize(width: 2*r, height: 2*r)
    }
    
    init(circleAt center:CGPoint, radius r: CGFloat) {
        self.init()
        origin = CGPoint(x: center.x - r, y: center.y - r)
        size = CGSize(width: 2 * r, height: 2 * r)
    }
    
    static let unitCircle = CGRect(x: -1, y: -1, width: 2, height: 2)
    
    var x: CGFloat { return origin.x }
    var y: CGFloat { return origin.y }
    var center: CGPoint { return CGPoint(x: midX, y: midY)}
}

enum Bond: Int, CustomStringConvertible {
    case bond, testBond, testSeed, testPrimary, primaryBestSoFar, primaryBest, seed, primary
    
    static let all = [ Bond.bond, .testBond, .primary, .testPrimary, .primaryBestSoFar, .primaryBest, .seed, .testSeed ] // in drawing order
    static let count = all.count
    
    var isTest : Bool {
        switch (self) {
        case .testBond, .testSeed, .testPrimary: return true
        default: return false
        }
    }
    
    var description: String {
        switch self {
        case .bond: return "Bond"
        case .testBond: return "Test bond"
        case .testSeed: return "Test seed"
        case .testPrimary: return "Test primary"
        case .primaryBestSoFar: return "Primary best so far"
        case .primaryBest: return "Primary best"
        case .seed: return "Seed"
        case .primary: return "Primary"
        }
    }
}

extension __CGContext {
    func makeArrow(from p: CGPoint, to q: CGPoint) {
        let s = sqrt(p.dist2(to: q))
        let a = atan2(q.y - p.y, q.x - p.x)
        saveGState()
        translateBy(x: p.x, y: p.y)
        scaleBy(x: s, y: s)
        rotate(by: a)
        move(to: CGPoint.zero)
        addLine(to: CGPoint(x: 0.4, y:0))
        addLine(to: CGPoint(x: 0.4, y:0.1))
        addLine(to: CGPoint(x: 0.7, y:0))
        addLine(to: CGPoint(x: 0.4, y:-0.1))
        addLine(to: CGPoint(x: 0.4, y:0))
        addLine(to: CGPoint(x: 1, y:0))
        restoreGState()
    }
    
    func addHexagon(at p: CGPoint, radius r: CGFloat) {
        move(to: CGPoint(x: p.x-r/2, y: p.y+2*IntPoint.sqrt32))
        addLine(to: CGPoint(x: p.x-r, y: p.y))
        addLine(to: CGPoint(x: p.x-r/2, y: p.y-2*IntPoint.sqrt32))
        addLine(to: CGPoint(x: p.x+r/2, y: p.y-2*IntPoint.sqrt32))
        addLine(to: CGPoint(x: p.x+r, y: p.y))
        addLine(to: CGPoint(x: p.x+r/2, y: p.y+2*IntPoint.sqrt32))
        closePath()
    }
    
}

extension Int {
    var bin: String {
        return String(self, radix:2)
    }
}

extension CGPoint {
    
    var norm2 : CGFloat { return x*x + y*y }
    
    private static let ε = CGFloat(1e-5)
    
    private static func solve2DEq(_ a: CGFloat, _ b: CGFloat, _ c: CGFloat, _ d: CGFloat, _ e: CGFloat, _ f: CGFloat) -> CGPoint? {
        let det = a*d-b*c
        
        guard abs(det) >= ε else { return nil }
        
        return CGPoint(x: (e*d-b*f)/det,
                       y: (a*f-e*c)/det)
    }
    
    static func centerOfCircle(englobing p: CGPoint, _ q: CGPoint, _ r: CGPoint) -> CGPoint? {
        let a = p.y - q.y
        let b = r.y - q.y
        let e = (r.x-p.x)/2
        let c = q.x-p.x
        let d = q.x-r.x
        let f = (r.y-p.y)/2
        
        guard let λμ = solve2DEq(a, b, c, d, e, f) else { return nil }
        
        return (p + q)/2 + λμ.x * CGPoint(x: a, y: c)
    }
    
    static func englobingCircle(of points: [CGPoint]) -> (center: CGPoint, radius: CGFloat) {
        // Rule out single point
        guard points.count >= 1 else { return (CGPoint.zero, 0) }
        guard points.count >= 2 else { return (points[0], 0) }
        
        // Look for diameter
        var allIn = false
        for i in 0..<points.count-1  {
            let p = points[i]
            for j in i+1..<points.count {
                let q = points[j]
                let c = (p + q)/2
                let r = (p - c).norm2
                
                allIn = true
                for k in 0..<points.count {
                    if (points[k] - c).norm2 > r + ε {
                        allIn = false
                        break
                    }
                }
                if allIn {
                    return (c, r)
                }
            }
        }
        
        // Now, look for non-flat triangle
        var bestCenter = CGPoint.zero
        var bestRadius = CGFloat.infinity
        for i in 0..<points.count-2 {
            let p = points[i]
            for j in (i+1)..<points.count-1 {
                let q = points[j]
                for k in (j+1)..<points.count {
                    let t = points[k]
                    guard let c = centerOfCircle(englobing: p, q, t) else { continue } // the three points are aligned
                    
                    let r = (p-c).norm2
                    
                    guard r < bestRadius else { continue } // radius higher than already found
                    
                    var allIn = true
                    for l in 0..<points.count {
                        if (points[l] - c).norm2 > r + ε {
                            allIn = false
                            break
                        }
                    }
                    
                    if allIn {
                        bestCenter = c
                        bestRadius = r
                    }
                }
            }
        }
        
        return (bestCenter, sqrt(bestRadius))
    }
    
    static func intPoints(in paths: [Path], from origin: IntPoint = IntPoint.zero) -> [IntPoint] {
        var points : Set<IntPoint> = [origin]
        for p in paths {
            var q = origin
            for d in p {
                q = q.neighbor(d)
                points.insert(q)
            }
        }
        return Array(points)
    }
    
    static func englobingCircle(of paths: [Path], from origin: IntPoint = IntPoint.zero) -> (center: CGPoint, radius: CGFloat) {
        let points : [CGPoint] = intPoints(in: paths, from: origin).map { $0.point }
        return englobingCircle(of: points)
    }
}

extension CGPoint {
    func floorCoordinates(inBasis vec1: IntPoint, _ vec2: IntPoint) -> (Int, Int) {
        let p = self
        let p1 = vec1.point
        let p2 = vec2.point
        
        let det = p1.x * p2.y - p1.y * p2.x
        let detI = p.x * p2.y - p.y * p2.x
        let i = Int(floor(detI / det))
        let detJ = p1.x * p.y - p1.y * p.x
        let j = Int(floor(detJ / det))
        
        return (i,j)
    }
    
    func closestGridPoint(withBasis vec1: IntPoint = Direction.E.intPoint, _ vec2: IntPoint = Direction.SW.intPoint) -> IntPoint {
        let p = self
        let (i,j) = p.floorCoordinates(inBasis: vec1, vec2)
        
        var best = IntPoint(x: i * vec1.x + j * vec2.x,
                            y: i * vec1.y + j * vec2.y) // to the left and above
        let p1 = best + vec2
        let p2 = best + vec1
        let p3 = p1 + vec1
        
        let otherCandidates = [ p1, p2, p3, ]
        for c in otherCandidates {
            if p.dist2(to: c.point) < p.dist2(to: best.point) {
                best = c
            }
        }
        
        return best
    }
    
    
}

extension Path {
    
    func boundingRect() -> CGRect {
        var (minX, maxX, minY, maxY) : (CGFloat,CGFloat,CGFloat,CGFloat) = (0, 0, 0, 0)
        var q = IntPoint.zero
        let path = self
        for d in path {
            q = q.neighbor(d)
            let p = q.point
            if p.x < minX { minX = p.x }
            if p.x > maxX { maxX = p.x }
            if p.y < minY { minY = p.y }
            if p.y > maxY { maxY = p.y }
        }
        return CGRect(x: minX, y: minY, width: maxX-minX, height: maxY-minY)
    }
}
