//
//  IntSequence.swift
//  CAOS Simulator
//
//  Created by Nikao on 22/06/2020.
//  Copyright © 2020 Nikao. All rights reserved.
//

import Foundation

enum IntSequence: CustomStringConvertible {
    case bt(Int,UInt), bta1([Int]), bta([Int],UInt), btr([Int],UInt), bts([IntSequence]), shift([Int],Int)
    
    var description: String {
        switch self {
        case let .bt(b,n): return "\(b)*\(n)"
        case let .bta1(a): return "\(a)"
        case let .bta(a,l): return "\(a)|\(l)"
        case let .btr(a,l): return "\(a)~\(l)"
        case let .bts(s): return s.description
        case let .shift(a,x): return "\(a)>>\(x)"
        }
    }
    
    func expand() -> [Int] {
        switch self {
        case let .bt(b,n): return [Int](repeating: b, count: Int(n))
        case let .bta1(a):
            return a
        case let .bta(a,l):
            var result = [Int]()
            for i in 0..<Int(l) {
                result.append(a[i%a.count])
            }
            return result
        case let .btr(a,l): // fill from the end of the sequence
            guard a.count > 0 else { return [] }
            var result = [Int]()
            let li = Int(l)
            let rl = a.count - (li % a.count)
            for i in 0..<li {
                result.append(a[(i+rl)%a.count])
            }
            return result
        case let .bts(s):
            var result = [Int]()
            for x in s {
                result.append(contentsOf: x.expand())
            }
            return result
        case let .shift(a,x):
            var result = [Int]()
            for i in 0..<a.count {
                result.append(a[(i+x)%a.count])
            }
            return result
        }
    }
}


extension Array {

    func loop(for l: Int) -> [Element] {
        guard l > 0 else { return [] }
        
        var result : [Element] = []
        result.reserveCapacity(l)
        for i in 0..<l {
            result.append(self[i%count])
        }
        return result
    }
    
    func backloop(for l: Int) -> [Element] {
        guard l > 0 else { return [] }
        
        var result : [Element] = []
        result.reserveCapacity(l)
        let li = (l % count) + count
        let rl = count - (li % count)
        for i in 0..<l {
            result.append(self[(i+rl)%count])
        }
        return result
    }
    
    func concat(with a: [Element]) -> [Element] {
        return self + a
    }
    
    func shift(by x : Int) -> [Element] {
        guard count > 0 else { return [] }
        
        let dx = (x % count) + count
        var result : [Element] = []
        for i in 0..<count {
            result.append(self[(i + dx) % count])
        }
        return result
    }
    
    mutating func add(_ a: [Element]) {
        self.append(contentsOf: a)
    }
    
}

extension Array where Element == Int {
    
    func mapto(_ a: Array) -> Array {
        var result = Array()
        for i in self {
            guard i < a.count else { return result }
            result.append(a[i])
        }
        return result
    }

}

indirect enum IntSequence2: CustomStringConvertible {
    case int(Int), array([Int]), loop(IntSequence2,for:Int), backloop(IntSequence2,for:Int), seq([IntSequence2]), shift(IntSequence2, by: Int), map(IntSequence2, [Int])
    
    var description: String {
        switch self {
        case let .int(i): return "\(i)"
        case let .array(a): return "\(a)"
        case let .loop(s, for: l): return "[\(s) for \(l)]"
        case let .backloop(s, for: l): return "[\(s) backward for \(l)]"
        case let .seq(s): return s.description
        case let .shift(s, by: x): return "\(s)>>\(x)"
        case let .map(s, m): return "[\(s):\(m)]"
        }
    }
    
    func expand() -> [Int] {
        switch self {
        case let .int(i): return [i]
        case let .array(a):
            return a
        case let .loop(s, for: l):
            let a = s.expand()
            guard a.count > 0 else { return [] }
            var result = [Int]()
            for i in 0..<Int(l) {
                result.append(a[i%a.count])
            }
            return result
        case let .backloop(s, for: l): // fill from the end of the sequence
            let a = s.expand()
            guard a.count > 0 else { return [] }
            var result = [Int]()
            let li = Int(l)
            let rl = a.count - (li % a.count)
            for i in 0..<li {
                result.append(a[(i+rl)%a.count])
            }
            return result
        case let .seq(s):
            var result = [Int]()
            for x in s {
                result.append(contentsOf: x.expand())
            }
            return result
        case let .shift(s, by: x):
            let a = s.expand()
            var result = [Int]()
            let dx = (x % a.count) + a.count
            for i in 0..<a.count {
                result.append(a[(i + dx) % a.count])
            }
            return result
        case let .map(s, m):
            let a = s.expand()
            return a.map { m[$0] }
        }
    }
}

enum NestedStringsAndInts: Decodable, CustomStringConvertible {
    
    case int(Int), str(String), array([NestedStringsAndInts])
    
    enum CodingKeys: String, CodingKey {
        case value = "value"
    }

    var description: String {
        switch self {
        case let .int(i): return "\(i)"
        case let .str(s): return "\"\(s)\""
        case let .array(a): return a.description
        }
    }
    
    var explicitDescription: String {
        switch self {
        case let .int(i): return ".int(\(i))"
        case let .str(s): return ".str(\"\(s)\")"
        case let .array(a): return ".array([\(a.map {$0.explicitDescription}.joined(separator: ", "))])"
        }
    }
    
    static func recurive_decoding(from container: inout UnkeyedDecodingContainer) -> NestedStringsAndInts {
        var result = [NestedStringsAndInts]()
        while !container.isAtEnd {
            if let i = try? container.decode(Int.self) {
                result.append(.int(i))
            } else if let v = try? container.decode(String.self) {
                result.append(.str(v))
            } else if var w = try? container.nestedUnkeyedContainer() {
                let aux = recurive_decoding(from: &w)
                result.append(aux)
            }
        }
        return .array(result)
    }

    
    init(from decoder: Decoder) throws {
        var container = try decoder.unkeyedContainer() // container(keyedBy: CodingKeys.self)
        var result = [NestedStringsAndInts]()
        while !container.isAtEnd {
            if let i = try? container.decode(Int.self) { //, forKey: CodingKeys.value) {
                result.append(.int(i))
            } else if let v = try? container.decode(String.self) { //, forKey: CodingKeys.value) {
                result.append(.str(v))
            } else if var v = try? container.nestedUnkeyedContainer() { //forKey: CodingKeys.value){
                result.append(NestedStringsAndInts.recurive_decoding(from: &v))
            } else {
                throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
            }
        }
        if result.count == 1 {
            self = result[0]
        } else {
            self = .array(result)
        }
    }
}

indirect enum CompactTranscript: CustomStringConvertible, Decodable {
    
    struct Keys {
        static let array = "array"
        static let unfold = "unfold"
        static let loop = "loop"
        static let declare = "declare"
        static let repeat_ = "repeat"
        static let mark = "mark"
    }
    
    case bt(String), array([CompactTranscript]), unfold(Int, Int, [CompactTranscript]), declare(String, CompactTranscript), repeat_(Int, Int, [CompactTranscript]), mark(String)
    
    var description: String {
        switch self {
        case let .bt(s): return s
        case let .array(a): return a.description
        case let .unfold(i,j,a): return "\(a){\(i)..<\(j)}"
        case let .repeat_(i,j,a): return "\(a){{\(i)..<\(j)}}"
        case let .declare(name, value): return "{\(name)=\(value)}"
        case let .mark(value): return "!\(value)!"
        }
    }
    
    func expand(from: Int = -1) -> (expansion: [String], marks: [(Int, String)]) {
        var variables = [String: (value:[String], marks:[(Int, String)])]()
        var marks = [(Int, String)]()
        var _from = from
        let expansion = expand(from: &_from, with: &variables, marks: &marks)
        print("VARIABLES:\n\(variables)")
        return (expansion, marks)
    }
    
    func expand(from: inout Int,
                with variables: inout [String:  (value:[String], marks:[(Int, String)])],
                marks: inout [(Int, String)]) -> [String] {
        switch self {
        case let .bt(s):
            if let (value, _marks) = variables[s] {
                for (l,s) in _marks {
                    marks.append((l+from, s))
                }
                from += value.count
                return value
            } else {
                from += 1
                return [s]
            }
        case let .array(a):
            return a.reduce([String]()) { $0 + $1.expand(from: &from, with: &variables, marks: &marks) }
        case let .unfold(i,j,a):
            var _from = -1
            var _marks = [(Int, String)]()
            let _a = CompactTranscript.array(a).expand(from: &_from, with: &variables, marks: &_marks)
            var result = [String](repeating: "", count: j-i)
            print("UNFOLD_MARKS: \(_marks)")
            for k in i..<j {
                from += 1
                let _k = (k%_a.count + _a.count)%_a.count
                result[k-i] = _a[_k]
                for (l,s) in _marks {
                    let _l = (l%_a.count + _a.count)%_a.count
                    if _l == _k { marks.append((from, s)) }
                    if _l >= _k { break }
                }
            }
            print("MARKS: \(marks)")
            return result
        case let .repeat_(i,j,a):
            var _a = [[String]]()
            var _marks = [[(Int, String)]]()
            for k in 0..<a.count {
                let t = a[k]
                var _from = -1
                _marks.append([(Int, String)]())
                _a.append(t.expand(from: &_from, with: &variables, marks: &(_marks[k])))
            }
            var result = [String]()
            for k in i..<j {
                let _k = (k%_a.count + _a.count)%_a.count
                result.append(contentsOf: _a[_k])
                for (l,s) in _marks[_k] {
                    marks.append((l+from,s))
                }
                from += _a[_k].count
            }
            return result
        case let .declare(name,value):
            var _from = -1
            var _marks = [(Int, String)]()
            let result = value.expand(from: &_from, with: &variables, marks: &_marks)
            variables[name] = (result, _marks)
            return []
        case let .mark(value):
            marks.append((from, value))
            return []
        }
    }
    
    // Decoding

    enum CodingKeys: String, CodingKey {
        case value = "value"
    }

    init(from decoder: Decoder) throws {
        let nestedStringsAndInts = try NestedStringsAndInts(from: decoder)
        self = try CompactTranscript(from: nestedStringsAndInts)
    }
    
    init(from nestedStringsAndInts: NestedStringsAndInts) throws {
        switch nestedStringsAndInts {
        case .int(_):
            throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
        case let .str(s):
            self = .bt(s)
        case let .array(a):
            switch a[0] {
            case let .str(k):
                if k == Keys.unfold {
                    switch (a[1], a[2]) {
                    case let (.int(i), .int(j)):
                        self = .unfold(i, j, try a[3...].map { try CompactTranscript(from: $0) } )
                        return
                    default:
                        throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
                    }
                } else if k == Keys.repeat_ {
                        switch (a[1], a[2]) {
                        case let (.int(i), .int(j)):
                            self = .repeat_(i, j, try a[3...].map { try CompactTranscript(from: $0) } )
                            return
                        default:
                            throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
                        }
                } else if k == Keys.loop {
                    switch a[1] {
                    case let .int(i):
                        self = .unfold(0, i, try a[2...].map { try CompactTranscript(from: $0) } )
                        return
                    default:
                        throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
                    }
                } else if k == Keys.declare {
                    switch a[1] {
                    case let .str(name):
                        print("Declare \(name): \(a[2...])")
                        let value = try CompactTranscript(from: NestedStringsAndInts.array(Array(a[2...])))
                        self = .declare(name, value)
                        return
                    default:
                        throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
                    }
                } else if k == Keys.mark {
                    if a.count == 1 {
                        self = .mark("")
                        return
                    }
                    switch a[1] {
                    case let .str(value):
                        print("Mark \(value)")
                        self = .mark(value)
                        return
                    default:
                        throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
                    }
                } else {
                    self = .array(try a.map { try CompactTranscript(from: $0)})
                    return
                }
            case .array(_):
                self = .array(try a.map { try CompactTranscript(from: $0)})
                return
            default:
                throw CodingErrors.parsingError("Cannot parse encoding as \(Self.self)")
            }
        }
    }
}

//    func encode(to encoder: Encoder) throws {
//        var container = encoder.container(keyedBy: CodingKeys.self)
//        switch self {
//        case let .bt(s): try container.encode(s, forKey: CodingKeys.bt)
//        case let .ids(a): try container.encode(a, forKey: CodingKeys.ids)
//        case let .array(a):
//            var array = container.nestedUnkeyedContainer(forKey: CodingKeys.array)
//            try array.encode(contentsOf: a)
//        case let .times(s, n):
//            try container.encode(n, forKey: CodingKeys.times)
//            try container.encode(s, forKey: CodingKeys.sequence)
//        case let .declare(v, s):
//            try container.encode(v, forKey: CodingKeys.declare)
//            try container.encode(s, forKey: CodingKeys.sequence)
//        case let .prefixOfInfiniteLoopOn(s,n):
//            try container.encode(n, forKey: CodingKeys.prefix)
//            try container.encode(s, forKey: CodingKeys.sequence)
//        case let .value(v):
//            try container.encode(v, forKey: CodingKeys.value)
//        }
//    }
