//
//  OpportunisticOritatamiSimulatorWithProofTrees
//  Scary Pacman 2
//
//  Created by Nikao on 16/3/18.
//  Copyright © 2018 Algoseum. All rights reserved.
//

import UIKit

class OpportunisticOritatamiSimulatorWithProofTrees { //} : OritatamiSimulator {    
    
    var grid: TriangularGridProtocol
    
    var oritatamiSystem: OritatamiSystem
    var beadTypeSet: BeadTypeSet { return oritatamiSystem.beadTypeSet }
    var beadTypeIdStream: [BeadTypeId] { return oritatamiSystem.beadTypeIdsSequence }
    var seedPath: [Direction] { return oritatamiSystem.seedPath }
    var delay: Int { return oritatamiSystem.delay }
    var beadStreamCount: Int { return oritatamiSystem.beadSequenceCount }
    
    var currentPosition = IntPoint.zero
    var currentIndex: Int = 0
    var mustStop = false
    
    weak var view: TriangularGridView? = nil
    var frameRate: Double = OpportunisticOritatamiSimulator.Defaults.frameRate
    var animated : Bool = true
    var follow : Bool = true
    
    var saveAsGif: Bool = false
    var shuffleGif: Bool = false
    var noLoopOnBest: Bool = true
    var gifFoldingStepRootFilename: String = "folding"
    var gifRunRootFilename: String = "OSsimulatorRun"
    var timedImageSequence : TimedImageSequence = []
    
    var currentBestFirstStep: [Direction] = []
    var currentBestNumberOfBonds: Int = 0
    
    required init(width w: Int = 100, height h: Int = 100,
                  oritatamiSystem os: OritatamiSystem) {
        grid = TriangularGrid(width: w, height: h)
        oritatamiSystem = os
        grid.beadTypeSet = os.beadTypeSet
        grid.rule = os.rule
    }

    var bestNumberOfBonds = 0
    var bestDir : [ Direction ] = []
    var isCurrentBestDir : Bool = false
    var bestPathsSoFar : [Path] = []
    var currentPath : Path = []
    var currentNumberOfBonds = 0

    @inline(__always) func beadTypeId(indexed j: Int) -> BeadTypeId {
        return oritatamiSystem.beadTypeId(indexed: j)
    }

    var pathBallLevelTree = PathBallLevelTree()
    var currentPreBallTree = [PathBallTree]()
    
    func initPathBallLevelTree() {
        pathBallLevelTree = PathBallLevelTree()
        currentPreBallTree = []
        
        let d0 = grid.occupiedNeighborDirection(from: currentPosition)
        let bt = oritatamiSystem.beadTypeId(indexed: currentIndex)
        
        var d = d0.cw
        while d != d0 {
            let v = currentPosition.neighbor(d)
            guard !grid.isOccupied(at: v) else { d = d.cw; continue }
            
            let nbBonds = grid.addBead(typeId: bt, from: currentPosition, to: d, isTest: true, isSeed: currentIndex < oritatamiSystem.seedPath.count)
            
            let pb = PathBall(origin: currentPosition)
            pb.insert(path: [d])
            pb.nbBonds = nbBonds
            
            let t = PathBallTree(pathBall: pb, parent: nil)
            currentPreBallTree.append(t)
            
            grid.removeBead(at: v)
            
            d = d.cw
        }
    }
    
    func extendCurrentPreBallTrees() {
        var newCurrentPreBallTrees = [PathBallTree]()
        
        var bestNbBonds = Int.min

        for t in currentPreBallTree {
            // Draw the common path
            let pb = t.pathBall
            // Draw common prefix of the paths in the balls
            let path0 = pb.paths[0]
            let (nbBondsPrefix, endPrefix) =
                grid.draw(path: path0,
                          beadTypeStream: oritatamiSystem.beadTypeId(indexed:),
                          isTesting: true,
                          isSeed: false,
                          from: pb.origin,
                          in: 0..<pb.commonPrefixLength)
           
            for path in pb.paths {
                // Draw the end of the path
                let (nbExtraBonds,endPath) =
                    grid.draw(path: path,
                          beadTypeStream: oritatamiSystem.beadTypeId(indexed:),
                          isTesting: true,
                          isSeed: false,
                          from: endPrefix,
                          in: pb.commonPrefixLength..<path.count)
                
                // Extend the path
                let noBondPathBall = PathBall(origin: pb.origin) // Ball where path with no new bond will be stored if any.
                noBondPathBall.transcriptIndex = pb.transcriptIndex
                noBondPathBall.nbBonds = pb.nbBonds

                // Look for the first direction
                let d0 = grid.occupiedNeighborDirection(from: endPath)
                let bt = oritatamiSystem.beadTypeId(indexed: currentIndex)
                
                var d = d0.cw
                while d != d0 {
                    let v = currentPosition.neighbor(d)
                    guard !grid.isOccupied(at: v) else { d = d.cw; continue }
                    
                    let nbNewBonds = grid.addBead(typeId: bt, from: currentPosition, to: d, isTest: true, isSeed: currentIndex < oritatamiSystem.seedPath.count)
                    let totNbBonds = nbBondsPrefix + nbExtraBonds + nbNewBonds
                    
                    let newPath = path + [d]
                    
                    if nbNewBonds == 0 {
                        noBondPathBall.insert(path: newPath)
                    } else { // create a new ball
                        let npb = PathBall(origin: currentPosition)
                        npb.insert(path: newPath)
                        npb.nbBonds = totNbBonds
                        
                        let t = PathBallTree(pathBall: npb, parent: t.parent)
                        newCurrentPreBallTrees.append(t)

                        if totNbBonds > bestNbBonds {
                            bestNbBonds = totNbBonds
                        }
                    }
                    
                    grid.removeBead(at: v)
                    
                    d = d.cw
                }
                
                if noBondPathBall.count > 0 {
                    let zt = PathBallTree(pathBall: noBondPathBall, parent: t.parent)
                    newCurrentPreBallTrees.append(zt)
                    if noBondPathBall.nbBonds > bestNbBonds {
                        bestNbBonds = noBondPathBall.nbBonds
                    }
                }
            }
        }
        
        // Tag the favorable
        // Tag the deadEnd
        currentPreBallTree = newCurrentPreBallTrees
    }
    
    func nextStep(index i: Int, lastPosition p: IntPoint) -> Int {
        guard i <= delay else { return 0 }
        
        let j = i + currentIndex
        guard j < beadStreamCount else { return -1 }
        
        let isLastStep = (i == delay || j == beadStreamCount-1)
        
        // Look for initial direction
        let d_init : Direction = .NE
        var d0 = d_init
        while !grid.isOccupied(at: p.neighbor(d0)) && d0 != d_init.ccw {
            d0 = d0.cw
        }
        
//        print("d0: \(d0) - i: \(i)")
        var d = d0.cw
        for _ in Direction.all {
            let q = p.neighbor(d)
            guard !grid.isOccupied(at: q) else { d = d.cw; continue }
            
            if i == 1 {
                currentPath = [d]
                isCurrentBestDir = false
            } else {
                currentPath.append(d)
//                print("\(i): \(currentPath)")
            }
            
            var nbNewBonds : Int = 0
            
            grid.synchronize {
                nbNewBonds = grid.addBead(typeId: beadTypeId(indexed: j), at: q, bond: .testBond)
                grid.setOutEdge(at: p, to: d)
                grid.setBond(at: p, to: d, value: .testPrimary)
            }

            currentNumberOfBonds += nbNewBonds
            
            if !isLastStep {
                let _ = nextStep(index: i+1, lastPosition: q)
            } else {
                var pathColor = TriangularGridView.Defaults.primaryTestBondColor
                var delay = (shuffleGif ?
                    OpportunisticOritatamiSimulator.Defaults.gifShuffleTestFrameRate :
                    OpportunisticOritatamiSimulator.Defaults.gifTestFrameRate)
                
//                print("Current path \(currentPath) from \(currentPosition)")
                
                if currentNumberOfBonds > bestNumberOfBonds {
                    bestNumberOfBonds =  currentNumberOfBonds
                    bestDir = []
                    bestPathsSoFar = []
//                    let v = view
//                    let info = "\(currentNumberOfBonds) bond\(currentNumberOfBonds >= 2 ? "s": "")"
//                    DispatchQueue.main.async {
//                        v?.controller?.show(info: info)
//                    }
                    if !shuffleGif {
                        pathColor = TriangularGridView.Defaults.primaryBestSoFarBondColor // __Color(white: 0.35, alpha: 1) // __Color(hue: CGFloat(currentNumberOfBonds)/5, saturation: 1, brightness: 1, alpha: 1)
                        delay *= 3
                    }
              }
                
                if currentNumberOfBonds >= bestNumberOfBonds {
                    isCurrentBestDir = true
//                    print("Added: \(currentPath)")
                    bestPathsSoFar.append(currentPath)
                    
//                    pathColor = __Color.lightBlue500 // __Color(white: 0.35, alpha: 1) // __Color(hue: CGFloat(currentNumberOfBonds)/5, saturation: 1, brightness: 1, alpha: 1)
//                    delay *= 2
                }

                let v = view
                let dp = currentPosition.point
                let r = CGFloat(delay)+2
                
                v?.primaryTestBondColor = pathColor
                DispatchQueue.main.async {
                    v?.setNeedsDisplay(inUnscaled:CGRect(circleAt: dp, radius:r))
                }
                
                if saveAsGif {
                    recordView(delay: delay, pathColor: pathColor)
                }
                
                if animated {
                    Thread.sleep(forTimeInterval: frameRate)
                }
                
            }
            
            grid.synchronize {
                grid.removeBead(at: q)
                DispatchQueue.main.async {
                    self.view?.setNeedsDisplay(inUnscaled:CGRect(circleAt: q.point, radius:1))
                }
            }
            currentPath.removeLast()
            currentNumberOfBonds -= nbNewBonds

            if i == 1 && isCurrentBestDir {
                bestDir.append(d)
            }
            
            d = d.cw
        }
        
        if i == 1 {
            // Best Paths
//           print("Best Dirs: \(bestDir)\nBest paths: \(bestPathsSoFar)\n")
            
            if shuffleGif && !timedImageSequence.isEmpty {
                let n = timedImageSequence.count
                for i in 0..<n-1 {
                    let j = n-i
                    let k = Int(arc4random())%j
                    if k < j-1 {
                        let v = timedImageSequence[k]
                        timedImageSequence[k] = timedImageSequence[j-1]
                        timedImageSequence[j-1] = v
                    }
                }
            }
            
            if animated && !noLoopOnBest {
                // Loop on best paths
                loopOnBestPath()
            }
            
            // Check for determinism
            if bestDir.count == 0 {
                print(" No extension!")
                return -1
            } else if bestDir.count == 1 {
                let d = bestDir[0]
//                print("Next direction is \(d)")
                // Make the step
                grid.synchronize {
                    currentPosition = p.neighbor(d)
                    currentIndex += 1
                    let _ = grid.addBead(typeId: beadTypeId(indexed:currentIndex), at: currentPosition, bond: .bond)
                    grid.setBeadIndex(at: currentPosition, value: currentIndex)
                    grid.setOutEdge(at: p, to: d)
                    grid.setBond(at: p, to: d, value: .primary)
                
                }
                
                let q = currentPosition.point
                DispatchQueue.main.async { [weak self] in
                    guard let me = self else { return }
                    if me.follow { me.view?.centerIfNeeded(at: me.currentPosition) }
                    me.view?.setNeedsDisplay(inUnscaled:CGRect(circleAt: q, radius:1))
                }

                
                return 0
            } else {
                print("Non-determinism! \(bestDir)")
                DispatchQueue.main.async { [weak self] in
                    guard let me = self else { return }
                    if me.follow { me.view?.centerIfNeeded(at: me.currentPosition) } }

                while true {
                    DispatchQueue.main.async { [weak self] in self?.view?.controller?.show(info:"Non-determinism! \(self!.bestDir)")
                    }
                    loopOnBestPath()
                }
                return -2
            }
        }
        
        return bestNumberOfBonds
    }
    
    func nextStep() -> Bool {
        bestDir = []
        bestNumberOfBonds = 0
        bestPathsSoFar = []
        currentPath = []
        let output = nextStep(index: 1, lastPosition: currentPosition)

        if saveAsGif {
            let _ = saveGif()
        }

        
        return output >= 0
    }

    func loopOnBestPath() {
        guard let v = view else { return }
        v.primaryTestBondColor = TriangularGridView.Defaults.primaryBestBondColor
        let delay = min(OpportunisticOritatamiSimulator.Defaults.gifLoopBestMaxFrameRate,
                        OpportunisticOritatamiSimulator.Defaults.gifLoopBestTotalFrameRate/CGFloat(bestPathsSoFar.count))
        for path in bestPathsSoFar {
//            print(path)
            grid.synchronize {
                draw(bestpath: path, erase: false)
            }
            DispatchQueue.main.sync { [weak self] in self?.view?.setNeedsDisplay() }
            if saveAsGif {
                recordView(delay: delay, pathColor: TriangularGridView.Defaults.primaryBestBondColor)
            }
            Thread.sleep(forTimeInterval: TimeInterval(delay))
            grid.synchronize {
                draw(bestpath: path, erase: true)
            }
        }
        v.primaryTestBondColor = TriangularGridView.Defaults.primaryTestBondColor
    }

    func draw(bestpath: Path, erase: Bool = false) {
        var p = currentPosition
        var j = currentIndex + 1
        if !erase {
            for d in bestpath {
                let q = p.neighbor(d)
                let _ = grid.addBead(typeId: beadTypeId(indexed: j), at: q, bond: .testBond)
                grid.setOutEdge(at: p, to: d)
                grid.setBond(at: p, to: d, value: .primaryBest)
                p = q
                j += 1
            }
        } else {
            for d in bestpath {
                p = p.neighbor(d)
                grid.removeBead(at: p)
            }
        }
    }
    

    
    func drawSeed(from p: IntPoint) {
//        print("Draw seed: \(seedPath) from \(p)")
        guard beadTypeIdStream.count > seedPath.count else { return }
        currentIndex = 0
        currentPosition = p
//        grid.synchronize {
            grid.clear()
            grid.pathStartingPoints.append(currentPosition)
        let _ = grid.addBead(typeId: beadTypeId(indexed: currentIndex), at: currentPosition, bond: .seed)
            grid.setBeadIndex(at: currentPosition, value: currentIndex)
            for d in seedPath {
                currentIndex += 1
                let p = currentPosition.neighbor(d)
                let _ = grid.addBead(typeId: beadTypeId(indexed: currentIndex), at: p, bond: .bond)
                grid.setBeadIndex(at: p, value: currentIndex)
                grid.setOutEdge(at: currentPosition, to: d)
                grid.setBond(at: currentPosition, to: d, value: .seed)
                currentPosition = p
            }
//        }
        print("Draw seed: Done, ends at \(currentPosition) current index \(currentIndex)")
    }
    
    
    // .MARK: GIF saving
    
    func saveGif(_ sequence: inout TimedImageSequence, to url: URL) -> Bool {
        let once : Bool = { print("URL: \(url)"); return true }()
        _ = once
        
        let result = sequence.saveAsGif(to: url, loopCount: 0)
        if (result) {
            sequence = []
        }
        return result
    }
    
    func saveGif(to url: URL) -> Bool {
        saveGif(&timedImageSequence, to: url)
    }

    
    private var printFileLocationOnce: Bool = true
    func saveGif() -> Bool {
        let filename = String(format: "%@-%04d.gif", gifFoldingStepRootFilename, currentIndex)
        guard let url = URL.documentURL(for: filename) else { return false }
        if printFileLocationOnce {
            print("Saving to \(url)...")
            printFileLocationOnce = false
        }
        let result = timedImageSequence.saveAsGif(to: url, loopCount: 0)
        if (result) {
            timedImageSequence = []
        }
//        print("Done \(result)")
        return result
    }
    
    func recordView(delay: CGFloat, pathColor: __Color) {
        guard let v = view else { return }
        var img : __Image? = nil
        DispatchQueue.main.sync {
            let h = v.controller?.infoLabel.isHidden ?? true
            CATransaction.setDisableActions(true)
            v.primaryTestBondColor = pathColor
            v.setNeedsDisplay()
            if shuffleGif { v.controller?.infoLabel.isHidden = true }
            img = __Image(view: v)
            if shuffleGif { v.controller?.infoLabel.isHidden = h }
            CATransaction.setDisableActions(false)
        }
        if let img = img {
            timedImageSequence.append((img, delay))
        }
    }
    
    func flushPendingGif() {
        print("Not implemented")
    }
}

//class PathTree<T> : CustomStringConvertible where T : CustomStringConvertible {
//    var value: T
//    var subtrees: [PathTree<T>] = []
//    weak var parent: PathTree<T>? = nil
//
//    init(value v: T) {
//        value = v
//    }
//
//    func addSubTree(_ t: PathTree<T>) {
//        if let p = t.parent {
//            p.removeSubTree(t)
//        }
//        subtrees.append(t)
//        t.parent = self
//    }
//
//    func removeSubTree(_ t: PathTree<T>) {
//        guard let p = t.parent, self === p else { return }
//        for i in 0..<subtrees.count {
//            if t === subtrees[i] {
//                subtrees.remove(at: i)
//                t.parent = nil
//                return
//            }
//        }
//    }
//
//    func _description(prefix: String = "") -> String {
//        var s = prefix + value.description + "\n"
//        let new = prefix + "  "
//        for t in subtrees {
//            s += t._description(prefix: new)
//        }
//        return s
//    }
//
//    var description: String {
//        return _description()
//    }
//}

