//
//  PrettyPrint.swift
//  Oritatami Builder
//
//  Created by Nikao on 23/09/2019.
//  Copyright © 2019 Nikao. All rights reserved.
//

import UIKit

extension OritatamiSystem {
    
    func createRuleMatrix() -> (beadTypeNames: [BeadType], attractionMatrix: [Bool]) {
        var usedBeadTypes = [BeadType]()
        for b1 in beadType {
            for b2 in beadType {
                if rule.attraction(between: b1, and: b2) {
                    usedBeadTypes.append(b1)
                    break
                }
            }
        }
        
        let n = usedBeadTypes.count
        print("Number of bead types used : \(n)");
        
        var attractionMatrix = [Bool](repeating: false, count: n*n)

        for i in 0..<n {
            for j in 0..<n {
                if rule.attraction(between: usedBeadTypes[i], and: usedBeadTypes[j]) {
                    attractionMatrix[i*n+j] = true
                    attractionMatrix[j*n+i] = true
                }
            }
        }
        
        return (usedBeadTypes, attractionMatrix)
    }

    func computeBeadTypesAttracted(to fromBeadTypes: [BeadType]) -> (left: [BeadType], right: [BeadType]) {
        
        let otherBeadTypes = beadType.filter() { !fromBeadTypes.contains($0) }
        
        var leftBeadTypes = [BeadType]()
        for b1 in fromBeadTypes {
            for b2 in beadType {
                if rule.attraction(between: b1, and: b2) {
                    leftBeadTypes.append(b1)
                    break
                }
            }
        }

        var rightBeadTypes = [BeadType]()
        for b1 in otherBeadTypes {
            for b2 in leftBeadTypes {
                if rule.attraction(between: b1, and: b2) {
                    rightBeadTypes.append(b1)
                    break
                }
            }
        }

        return (leftBeadTypes, rightBeadTypes)
    }
    
    func createPartialRuleMatrix(from leftBeadTypes: [BeadType]) -> (beadType: [BeadType], left: Int, right: Int, attractionMatrix: [Bool]) {

        let (left, right) = computeBeadTypesAttracted(to: leftBeadTypes)
        
        let l = left.count
        let r = right.count
        let n = l + r
        
        let usedBeadTypes = left + right
        var attractionMatrix = [Bool](repeating: false, count: n*n)

        for i in 0..<n {
            for j in 0..<n {
                if rule.attraction(between: usedBeadTypes[i], and: usedBeadTypes[j]) {
                    attractionMatrix[i*n+j] = true
                    attractionMatrix[j*n+i] = true
                }
            }
        }
        
        return (usedBeadTypes, l, r, attractionMatrix)
    }
 
    static let arcEnCiel = [__Color.yellow900,
                            .green900,
                            .blue900,
                            .pink900,
                            .amber900,
                            .lime900,
                            .lightBlue900,
                            .purple900,
                            .deepOrange900,
                            .lightGreen900,
                            .cyan900,
                            .deepPurple900,
                            .orange900,
                            .teal900,
                            .indigo900,
                            .blueGrey900,
                            .brown900,
                            .redA700 ]
    
    func saveRule(attractionMatrix: [Bool], beadTypes: [BeadType], left: Int, right: Int, side delta: CGFloat, font: __Font, heart: Bool, arcEnCiel: Bool = false, fileName: String) {
        
        func point(for i: Int) -> CGPoint {
            return i < left ?
                CGPoint(x:0,
                        y: CGFloat(left - 1 - 2 * i) * delta)
                : CGPoint(x: CGFloat(left + right) * delta,
                          y: CGFloat(right - 1 - 2 * (i - left)) * delta)
        }

        func intersection(for i_: Int, and j_: Int) -> CGPoint {
            let i = min(i_, j_)
            let j = max(i_, j_)
            let pi = point(for: i)
            let pj = point(for: j)
            if j < left {
                let dy = (pi.y - pj.y)/2
                return pi + CGPoint(x: dy, y: -dy)
            } else if i < left {
                let s = (pj + pi)/2
                let d = (pj - pi)/2
                if abs(s.y + d.x) <= abs(s.y - d.x) {
                    return s + CGPoint(x: d.y, y: d.x)
                } else {
                    return s - CGPoint(x: d.y, y: d.x)
                }
            } else {
                let dy = (pi.y - pj.y)/2
                return pi - CGPoint(x: dy, y: dy)
            }
        }
        
        
        let n = left + right
        let radius = heart ? delta/1.4 : delta/2
        let lineWidth = delta/7
        var nameAttr : [NSAttributedString.Key: Any] = [ .font: font ]
        let outlineAttr : [NSAttributedString.Key: Any] = [
            .font:font,
            .strokeColor: __Color.white,
            .strokeWidth: font.pointSize*2,
        ]

        var names = [NSAttributedString]()
        var namesOutlined = [NSAttributedString]()
        var nameSize = CGSize.zero
        for i in 0..<n {
            let c = arcEnCiel ? OritatamiSystem.arcEnCiel[(i/4)%OritatamiSystem.arcEnCiel.count] : beadTypes[i].color
            nameAttr[.foregroundColor] = c
            let name = beadTypes[i].name
            let attributedName = NSAttributedString(string: name, attributes: nameAttr)
            names.append(attributedName)
            let attributedNameOutlined = NSAttributedString(string: name, attributes: outlineAttr)
            namesOutlined.append(attributedNameOutlined)

            let size1 = attributedName.size()
            let size2 = attributedNameOutlined.size()

            if size1.width > nameSize.width {
                nameSize.width = size1.width
            }
            if size2.width > nameSize.width {
                nameSize.width = size2.width
            }
        }

        let namesPerRow = 1+ceil((font.ascender-font.descender)/2/delta)
        
        let beadTypeNameWidth = namesPerRow * (nameSize.width + delta) + 2 * delta
        let width = delta * CGFloat(n) + 2 * beadTypeNameWidth
        let height = 2 * (1 + CGFloat(max(left, right))) * delta
            + 2 * max(0,nameSize.height - delta)

        var hMin = -height / 2 + delta
        var hMax = height / 2 - delta
        
        for i in 0..<n {
            for j in i..<n {
                if attractionMatrix[i*n+j] {
                    let q = intersection(for: i, and: j)
                    if q.y < hMin { hMin = q.y }
                    if q.y > hMax { hMax = q.y }
                }
            }
        }
        hMin -= delta;
        hMax += delta;
        
        let heartPDFDocument = CGPDFDocument(Bundle.main.url(forResource: "Heart-white", withExtension: "pdf")! as CFURL)!
        let heartPDFPage = heartPDFDocument.page(at: 1)!
        let heartFrame = heartPDFPage.getBoxRect(.mediaBox)
        
        let scale = (2 * radius) / max(heartFrame.size.width, heartFrame.size.height)
        
        let fileManager = FileManager.default
        do {
            let documentDirectory = try fileManager.url(for: .documentDirectory, in: .userDomainMask, appropriateFor:nil, create:false)
            let fileURL = documentDirectory.appendingPathComponent(fileName + ".pdf")
            print("Saving to \(fileURL)")

            let mediaBox = CGRect(x: 0, y: 0, width: width, height: hMax-hMin)
            __GraphicsBeginPDFContextToFile(fileURL.path, mediaBox, nil)
            __GraphicsBeginPDFPage()
            let ctx = __GraphicsGetCurrentContext()!
            ctx.translateBy(x: beadTypeNameWidth, y: -hMin)
        
            ctx.setLineWidth(lineWidth)
            var x : CGFloat = 0
            
            for i in 0..<n {
                let p = point(for: i)
                var pMax = p
                var pMin = p
                for j in 0..<n {
                    if attractionMatrix[i*n+j] {
                        let q = intersection(for: i, and: j)
                        if q.y < pMin.y { pMin = q }
                        if q.y > pMax.y { pMax = q }
                    }
                }
                
                let c = arcEnCiel ? OritatamiSystem.arcEnCiel[(i/4)%OritatamiSystem.arcEnCiel.count] : beadTypes[i].color
                ctx.setStrokeColor(c.cgColor)
                switch (i%4) {
                case 0, 2:
                    ctx.setLineDash(phase: 0, lengths: [])
                    break
                case 1:
                    ctx.setLineDash(phase: 0, lengths: [ 4 * lineWidth, 4 * lineWidth ])
                    break
                default: // case 3:
                    ctx.setLineDash(phase: 0, lengths: [ lineWidth, 3 * lineWidth ])
                    break
                }
                ctx.addLines(between: [ pMax, p, pMin ])
                
                if i < left {
                    let nameSize = names[i].size()
                    let q = p
                        + CGPoint(x: -nameSize.width + x - 2 * delta ,
                                  y: -delta / 2 - (nameSize.height - font.ascender + font.descender) * 2)
                    ctx.addLines(between: [
                        p + CGPoint(x: x-2*delta, y:delta),
                        p + CGPoint(x: x-delta, y: delta),
                        p + CGPoint(x: x-delta, y: 0),
                        p,
                    ])
                    ctx.strokePath()
                    ctx.setLineDash(phase: 0, lengths: [])
                    namesOutlined[i].draw(at: q)
                    names[i].draw(at: q)
                
                    if i%Int(namesPerRow) != Int(namesPerRow)-1 {
                        x -= nameSize.width + delta
                    } else {
                        x = 0
                    }
                } else {
                    if (i == left) {
                        x = 0
                    }
                    let nameSize = names[i].size()
                    let q = p + CGPoint(x: x+2*delta, y: -delta/2-(nameSize.height-font.ascender+font.descender)*2)
                
                    ctx.addLines(between: [
                        p + CGPoint(x: x+2*delta, y: delta),
                        p + CGPoint(x: x+delta, y: delta),
                        p + CGPoint(x: x+delta, y: 0),
                        p,
                    ])
                    ctx.strokePath()
                    ctx.setLineDash(phase: 0, lengths: [])
                    namesOutlined[i].draw(at: q)
                    names[i].draw(at: q)
                
                    if (i-left)%Int(namesPerRow) != Int(namesPerRow)-1 {
                        x += nameSize.width + delta
                    } else {
                        x = 0
                    }
            }
        }
        
            ctx.setFillColor(__Color.black.cgColor)
            ctx.setStrokeColor(__Color.white.cgColor)

            for i in 0..<n {
                for j in i..<n {
                    if attractionMatrix[i*n+j] {
                        let p = intersection(for: i, and: j)
                    
                        if !heart {
                            let r = p + CGRect(x: -radius, y: -radius,
                                               width: 2*radius, height: 2*radius)
                            ctx.strokeEllipse(in: r)
                            ctx.fillEllipse(in: r)
                        } else {
                            ctx.saveGState()
                            ctx.translateBy(x: p.x, y: p.y)
                            ctx.scaleBy(x: scale, y: -scale)
                            ctx.translateBy(x: -heartFrame.size.width/2,
                                y: -heartFrame.size.height/2)
                            ctx.drawPDFPage(heartPDFPage)
                            ctx.restoreGState()
                    }
                }
            }
        }

        print("Closing Rule PDF")
            ctx.endPage()
            ctx.closePDF()
        } catch {
            print(error)
        }
    }

    func saveRule(rootFileName: String) {
        // Save general matrix
        let (b,m) = createRuleMatrix()
        saveRule(attractionMatrix: m, beadTypes: b, left: b.count-b.count/2, right: b.count/2, side: CGFloat(20*72)/2.5/CGFloat(b.count), font: __Font.systemFont(ofSize: 6), heart: false, fileName: rootFileName)
        
        // Save one matrix by category
        for c in beadTypeSet.category {
            let bt = beadTypeSet.beadType.filter() { $0.categoryId == c.id }
            let (b,l,r,m) = createPartialRuleMatrix(from: bt)
            let filename = rootFileName + "-" + c.name
            let delta = CGFloat(10)
            let font = __Font.systemFont(ofSize: 2.25*delta)
            saveRule(attractionMatrix: m, beadTypes: b, left: l, right: r, side: delta, font: font, heart: false, fileName: filename)
        }
        
        print(rule.prettyDescription(with: beadTypeSet))
    }

    func saveSubRule(for subset: Set<BeadTypeId>, filename: String) {
        let bt = beadTypeSet.beadType.filter() { subset.contains($0.id) }
        print("beadtypes: \(bt)")
        let (b,l,r,m) = createPartialRuleMatrix(from: bt)
        let delta = CGFloat(10)
        let font = __Font.systemFont(ofSize: 2.25*delta)
        saveRule(attractionMatrix: m, beadTypes: b, left: l, right: r, side: delta, font: font, heart: false, fileName: filename)
    }
    
}
