//
//  TriangularGridDictionary.swift
//  Simple OS Simulator
//
//  Created by Nikao on 27/05/2021.
//  Copyright © 2021 Nikao. All rights reserved.
//

import Foundation
import QuartzCore

import QuartzCore

struct DataNode {
    var beadTypeId: BeadTypeId? = nil
    var beadIndex: Int? = nil
    var bond: [Bond?] = Array(repeating: nil, count: Direction.halfCount)
    var outEdge: Direction? = nil
    var info: GridInfo? = nil
}

class TriangularGridDictionary : TriangularGridProtocol {
    
    // MARK: sync queue
    static let queueName = "gridDictQueue"
    let queue = DispatchQueue(label: TriangularGrid.queueName, qos: .userInitiated, attributes: DispatchQueue.Attributes()) // sequential queue
    
    func currentQueueName() -> String? {
        let name = __dispatch_queue_get_label(nil)
        return String(cString: name, encoding: .utf8)
    }
    
    func synchronize(closure: () -> ()) {
        queue.sync { closure() }
    }
    
    func synchronizeIfNotAlreadySynchronized(closure: () -> Void) {
        if currentQueueName() == TriangularGrid.queueName {
            closure()
        } else {
            queue.sync { closure() }
        }
    }

    // MARK: dictionnary
    var dict_bitsize = Defaults.dict_bitsize {
        didSet {
            dict_size = 1<<dict_bitsize
            dict_mask = dict_size-1
        }
    }
    var dict_size = 0
    var dict_mask = 0
    
    var dict_of_dict: [IntPoint : [IntPoint : DataNode]] = [:]

    // MARK: flush inaccessible beads
    
    var last_flush_dict_count : [IntPoint: Int] = [:]
    
    // MARK: cache
    var cache_cell_bitsize = Defaults.cache_cell_bitsize {
        didSet {
            reset_cache_with_cache_cell_bitsize()
        }
    }
    
    var cache_cell_size = 0 // Il faut 2^cache_bitsize ≥ δ+1
    var cache_cell_mask = 0
    
    var cache_bitsize = 0
    var cache_size = 0
    var cache_mask = 0
    
    var cache = [DataNode?]()
    var cache_origin: IntPoint = IntPoint.zero // coordonées réelles (par pas de cache_cell_size)
    
    // Compressed path
    var compressed_path = CompressedPath()
    
    // Other

    var minX: Int = Int.min
    var maxX: Int = Int.max
    var minY: Int = Int.min
    var maxY: Int = Int.max
    
    var pathStartingPoints: [IntPoint] = []
    var beadTypeSet: BeadTypeSet? = nil
    var rule: OritatamiRule? = nil
        
    init() {
        reset_dict_with_dict_bitsize()
        reset_cache_with_cache_cell_bitsize()
    }
    
    func reset_dict_with_dict_bitsize() {
        print("Reset Dictionnary")
        dict_size = 1<<dict_bitsize
        dict_mask = dict_size-1
        dict_of_dict = [:]
    }
    
    func reset_cache_with_cache_cell_bitsize() {
        print("Reset Cache")
        cache_cell_size = 1<<cache_cell_bitsize
        cache_cell_mask = cache_cell_size - 1
        cache_bitsize = cache_cell_bitsize+2
        cache_size = 1<<cache_bitsize
        cache_mask = cache_size - 1

        let cs = cache_size
        cache = Array(repeating: nil, count: cs * cs)
    }
    
    func clear() {
        dict_of_dict = [:]
        pathStartingPoints = []
        
        for i in 0..<cache.count {
            cache[i] = nil
        }
        cache_origin = IntPoint.zero
    }
    
    // MARK: dictionary access
    
    func data_in_dict_of_dict(at p: IntPoint) -> DataNode? {
        // works for negative numbers as well
        let (db, dm) = (dict_bitsize, dict_mask)
        let x1 = p.x>>db
        let y1 = p.y>>db

        guard let dict = dict_of_dict[IntPoint(x: x1, y: y1)] else { return nil }

        let x0 = p.x&dm
        let y0 = p.y&dm
        return dict[IntPoint(x: x0, y: y0)]
    }
    
    // MARK: cache init
    
    func fill_cache_cells() {
        let ccs = cache_cell_size
        // Fill the cache cells
        for y in 0..<4 {
            for x in 0..<4 {
                fill_cache_cell(at: IntPoint(x: cache_origin.x + x*ccs, y: cache_origin.y + y*ccs))
            }
        }
    }
    
    func fill_cache_cell(at top_left: IntPoint) {
        let (cb, cm) = (cache_bitsize, cache_mask)
        let (ccs, ccm) = (cache_cell_size, cache_cell_mask)
        let (db, dm) = (dict_bitsize, dict_mask)
        
        assert(db >= cb, "Cache is larger than dictionary squares")
        assert(top_left.x&ccm == 0 && top_left.y&ccm == 0, "\(top_left) is not the topleft corner of a cache cell")
        
        // Get the dictionnary containing the cache cell
        let dict = dict_of_dict[IntPoint(x: top_left.x>>db, y: top_left.y>>db)] ?? [:]
        
        // Fill the cache cell
        let tlx = top_left.x
        let tly = top_left.y
        for y in tly..<tly+ccs {
            for x in tlx..<tlx+ccs {
                cache[(x&cm)|((y&cm)<<cb)] = dict[IntPoint(x: x&dm, y: y&dm)]
            }
        }
    }
    
    func init_cache(at o: IntPoint) {
        // o will be placed in the center top-left cell
        assert(dict_bitsize >= cache_bitsize, "Cache is larger than dictionary squares")
        
        // /!\ DOES NOT Check for uncomitted points -> Push committed points first !

        let (ccs, ccm) = (cache_cell_size, cache_cell_mask)
        
        // Set the cache origin
        let ox = o.x&(~ccm) // round to the origin of the cache_cell containing o
        let oy = o.y&(~ccm) // round to the origin of the cache_cell containing o
        cache_origin = IntPoint(x: ox - ccs, y: oy - ccs)
        
        fill_cache_cells()
    }
    
    // MARK: mutate dataNodes
    
    func mutate(dataNode: inout DataNode, using mutating_block: (inout DataNode) -> ()) {
        mutating_block(&dataNode)
    }
    
    func mutate(dataNode_dict: inout [IntPoint:DataNode], using mutating_block: (inout [IntPoint:DataNode]) -> ()) {
        mutating_block(&dataNode_dict)
    }
    
    func mutate(dataNode_dict_of_dict: inout [IntPoint:[IntPoint:DataNode]], using mutating_block: (inout [IntPoint:[IntPoint:DataNode]]) -> ()) {
        mutating_block(&dataNode_dict_of_dict)
    }
    
    // MARK: cache test
    
    func cache_contains(_ p: IntPoint) -> Bool {
        let co = cache_origin
        let cs = cache_size
        return p.x >= co.x && p.x < co.x + cs
            &&  p.y >= co.y && p.y < co.y + cs
    }
    
    func cache_center_contains(_ p: IntPoint) -> Bool {
        let co = cache_origin
        let ccs = cache_size
        let ccs3 = (ccs<<1)|ccs
        return p.x >= co.x + ccs
            &&  p.x < co.x + ccs3
            &&  p.y >= co.y + ccs
            &&  p.y < co.y + ccs3
    }
    
    // MARK: shift cache
    
    func shift_cache(_ dir: Direction) {
        if auto_push {
            push_committed_points()
        }
        
        let ccs = cache_cell_size
        var cells_to_update : Array<(Int,Int)>
        switch dir {
        case .E:
            cache_origin.x += ccs
            cells_to_update = [(3,0), (3,1), (3,2), (3,3)]
        case .SE:
            cache_origin.x += ccs
            cache_origin.y += ccs
            cells_to_update = [(3,0), (3,1), (3,2), (3,3), (2,3), (1,3), (0,3)]
        case .SW:
            cache_origin.y += ccs
            cells_to_update = [(0,3), (1,3), (2,3), (3,3)]
        case .W:
            cache_origin.x -= ccs
            cells_to_update = [(0,0), (0,1), (0,2), (0,3)]
        case .NW:
            cache_origin.x -= ccs
            cache_origin.y -= ccs
            cells_to_update = [(0,0), (0,1), (0,2), (0,3), (1,0), (2,0), (3,0)]
        case .NE:
            cache_origin.y -= ccs
            cells_to_update = [(0,0), (1,0), (2,0), (3,0)]
        }
        
        for (x,y) in cells_to_update {
            fill_cache_cell(at: IntPoint(x: cache_origin.x + x*ccs, y: cache_origin.y + y*ccs))
        }
    }
    
    func shift_cache(to_include_in_center p: IntPoint) {
        // Does not push the changes, changes will be lost!
        let ccs = cache_cell_size
        let ccs3 = ccs<<1|ccs
        while p.x < cache_origin.x + ccs {
//            print("shift cache W")
            shift_cache(.W)
        }
        while p.x >= cache_origin.x + ccs3 {
//            print("shift cache E")
            shift_cache(.E)
        }
        while p.y < cache_origin.y + ccs {
//            print("shift cache NE")
            shift_cache(.NE)
        }
        while p.y >= cache_origin.y + ccs3 {
//            print("shift cache SW")
            shift_cache(.SW)
        }
    }
    
    // MARK: cache push
    
    var auto_push = Defaults.auto_push
    var push_includes_neighbors = Defaults.push_includes_neighbors
    var committed_points : Set<IntPoint> = Set()
    
    func commit(point p: IntPoint) {
        // /!\ assumes that p is in the cache, if not then non-determinism will occur
        committed_points.insert(p)
    }
    
    func push_committed_points() {
        if committed_points.count > 0 {
            push_cache(from: committed_points, push_includes_neighbors: push_includes_neighbors)
            committed_points.removeAll()
        }
    }

    
    func push_cache(from point_set: Set<IntPoint>, push_includes_neighbors: Bool = Defaults.push_includes_neighbors) {
        // assume the points in set are in the cache
        
        // Get the dictionnary
        let (cb, cm) = (cache_bitsize, cache_mask)
        let (db, dm) = (dict_bitsize, dict_mask)
        
        // get the points coordinate in dict
        var dict_point_set = point_set

        // add their neighbors
        if push_includes_neighbors {
            for p in point_set {
                for d in Direction.all {
                    dict_point_set.insert(p.neighbor(d))
                }
            }
        }
        
        // push the cache to the dictionnary
        for p in dict_point_set {
            let dx = p.x>>db
            let dy = p.y>>db
            let dp = IntPoint(x: dx, y: dy)
            if dict_of_dict[dp] == nil {
                dict_of_dict[dp] = [:]
            }
            let pd = IntPoint(x: p.x&dm, y: p.y&dm)
            let qx = p.x&cm
            let qy = p.y&cm
            dict_of_dict[dp]![pd] = cache[qx|(qy<<cb)]
        }
    }

    // MARK: read data

    func getData(at p: IntPoint, readFromCacheWithoutChecking: Bool = false) -> DataNode? {
        let (cb,cm) = (cache_bitsize, cache_mask)
        if readFromCacheWithoutChecking || cache_contains(p) {
//            print("read from cache")
            return cache[(p.x&cm)|((p.y&cm)<<cb)]
        }
        
        let (db, dm) = (dict_bitsize, dict_mask)
        let dx = p.x>>db
        let dy = p.y>>db
        if let dict = dict_of_dict[IntPoint(x: dx, y: dy)] {
            return dict[IntPoint(x: p.x&dm, y: p.y&dm)]
        }
        return nil
    }
    
    // MARK: write data
    
    func changeData(at p: IntPoint, writeToCacheWithoutChecking: Bool = false,
                    using block: (DataNode?) -> DataNode?) {
        // Write to cache only if p is in the cache or forced to use the cache, otherwise write in the dict
        let (cb,cm) = (cache_bitsize, cache_mask)
        if writeToCacheWithoutChecking || cache_contains(p) {
            let i = (p.x&cm)|((p.y&cm)<<cb)
            cache[i] = block(cache[i])
        }
        
        let (db, dm) = (dict_bitsize, dict_mask)
        let dx = p.x>>db
        let dy = p.y>>db
        let pdd = IntPoint(x: dx, y: dy)
        if dict_of_dict[pdd] == nil {
            dict_of_dict[pdd] = [:]
        }
        let pd = IntPoint(x: p.x&dm, y: p.y&dm)
        dict_of_dict[pdd]![pd] = block(dict_of_dict[pdd]![pd])
    }

    func mutateData(at p: IntPoint, writeToCacheWithoutChecking: Bool = false,
                    using mutating_block: (inout DataNode) -> ()) {
        // Write to cache only if p is in the cache or forced to use the cache, otherwise write in the dict
        let (cb,cm) = (cache_bitsize, cache_mask)
        if writeToCacheWithoutChecking || cache_contains(p) {
            let i = (p.x&cm)|((p.y&cm)<<cb)
            if cache[i] == nil {
                cache[i] = DataNode()
            }
            mutate(dataNode: &(cache[i]!), using: mutating_block)
        }
        
        let (db, dm) = (dict_bitsize, dict_mask)
        let dx = p.x>>db
        let dy = p.y>>db
        let pdd = IntPoint(x: dx, y: dy)
        if dict_of_dict[pdd] == nil {
            dict_of_dict[pdd] = [:]
        }
        let pd = IntPoint(x: p.x&dm, y: p.y&dm)
        if dict_of_dict[pdd]![pd] == nil {
            dict_of_dict[pdd]![pd] = DataNode()
        }
        mutate(dataNode: &(dict_of_dict[pdd]![pd]!), using: mutating_block)
    }
    
    // MARK: triangular grid protocol

    func isOccupied(at p: IntPoint) -> Bool {
        return isOccupied(at: p, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }

    func beadTypeId(at p: IntPoint) -> BeadTypeId? {
        return beadTypeId(at: p, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func setBeadTypeId(at p: IntPoint, value: BeadTypeId?) {
        setBeadTypeId(at: p, value: value, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }

    func beadIndex(at p: IntPoint) -> Int? {
        return beadIndex(at: p, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func setBeadIndex(at p: IntPoint, value: Int?) {
        setBeadIndex(at: p, value: value, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func bond(at p: IntPoint, direction d: Direction) -> Bond? {
        return bond(at: p, direction: d, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }

    func setBond(at p: IntPoint, to d: Direction, value: Bond?) {
        setBond(at: p, to: d, value: value, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func outEdge(at p: IntPoint) -> Direction? {
        return outEdge(at: p, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func setOutEdge(at p: IntPoint, to d: Direction?) {
        setOutEdge(at: p, to: d, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func addBead(typeId: BeadTypeId, from p: IntPoint, to: Direction, isTest: Bool, isSeed: Bool) -> Int {
        addBead(typeId: typeId, from: p, to: to, isTest: isTest, isSeed: isSeed, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func addBead(typeId: BeadTypeId, at p: IntPoint, bond: Bond) -> Int {
        addBead(typeId: typeId, at: p, bond: bond, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func removeBead(at p: IntPoint) {
        removeBead(at: p, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func draw(path: Path, beadTypeStream: (Int) -> BeadTypeId, isTesting: Bool, isSeed: Bool, from p: IntPoint, in range: Range<Int>?) -> (nbBonds: Int, endPoint: IntPoint) {
    
        draw(path: path, beadTypeStream: beadTypeStream, isTesting: isTesting, isSeed: isSeed, from: p, in: range, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func draw(conformation: Conformation, isTesting: Bool, isSeed: Bool, from p: IntPoint) -> (nbBonds: Int, endPoint: IntPoint) {
        draw(conformation: conformation, isTesting: isTesting, isSeed: isSeed, from: p, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func erase(path: Path, from p: IntPoint, shortenBy start: Int) {
        erase(path: path, from: p, shortenBy: start, writeToCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }
    
    func occupiedNeighborDirection(from p: IntPoint) -> Direction {
        return occupiedNeighborDirection(from: p, readFromCacheWithoutChecking: Defaults.read_write_from_cache_by_default)
    }

    
    // MARK: implementation
    
    func isOccupied(at p: IntPoint,
                    readFromCacheWithoutChecking rfcwc: Bool) -> Bool {
        if let data = getData(at: p, readFromCacheWithoutChecking: rfcwc) {
            return data.beadTypeId != nil
        }
        return false
    }
    
    func beadTypeId(at p: IntPoint, readFromCacheWithoutChecking rfcwc: Bool) -> BeadTypeId? {
        if let data = getData(at: p, readFromCacheWithoutChecking: rfcwc) {
            return data.beadTypeId
        }
        return nil
    }

    func setBeadTypeId(at p: IntPoint, value: BeadTypeId?,
                       writeToCacheWithoutChecking wtcwc: Bool) {
        mutateData(at: p, writeToCacheWithoutChecking: wtcwc) { $0.beadTypeId = value }
    }
    
    func beadIndex(at p: IntPoint,
                   readFromCacheWithoutChecking rfcwc: Bool) -> Int? {
        if let data = getData(at: p, readFromCacheWithoutChecking: rfcwc) {
            return data.beadIndex
        }
        return nil
    }

    func setBeadIndex(at p: IntPoint, value: Int?,
                      writeToCacheWithoutChecking wtcwc: Bool) {
        mutateData(at: p, writeToCacheWithoutChecking: wtcwc) { $0.beadIndex = value }
    }

    func bond(at p: IntPoint, direction d: Direction,
              readFromCacheWithoutChecking rfcwc: Bool) -> Bond? {
        if d.rawValue < Direction.halfCount {
            if let data = getData(at: p, readFromCacheWithoutChecking: rfcwc) {
                return data.bond[d.rawValue]
            }
            return nil
        } else {
            if let data = getData(at: p.neighbor(d), readFromCacheWithoutChecking: rfcwc) {
                return data.bond[(-d).rawValue]
            }
            return nil
        }
        return nil
    }
    
    func setBond(at p: IntPoint, to d: Direction, value: Bond?,
                 writeToCacheWithoutChecking wtcwc: Bool) {
        if d.rawValue < Direction.halfCount {
            mutateData(at: p, writeToCacheWithoutChecking: wtcwc) { $0.bond[d.rawValue] = value }
        } else {
            mutateData(at: p.neighbor(d), writeToCacheWithoutChecking: wtcwc) { $0.bond[(-d).rawValue] = value }
        }
    }
    
    func outEdge(at p: IntPoint,
                 readFromCacheWithoutChecking rfcwc: Bool) -> Direction? {
        if let data = getData(at: p, readFromCacheWithoutChecking: rfcwc) {
            return data.outEdge
        }
        return nil
    }
    
    func setOutEdge(at p: IntPoint, to d: Direction?,
                    writeToCacheWithoutChecking wtcwc: Bool) {
        mutateData(at: p, writeToCacheWithoutChecking: wtcwc) {
            $0.outEdge = d
        }

    }
    
    func addBead(typeId: BeadTypeId, from p: IntPoint, to: Direction, isTest: Bool, isSeed: Bool,
                 writeToCacheWithoutChecking wtcwc: Bool) -> Int {
        let q = p.neighbor(to)
        var bondCount = 0
        setBeadTypeId(at: q, value: typeId, writeToCacheWithoutChecking: wtcwc)
        let bond : Bond = isTest ? .testBond : .bond
        let opp = to.opposite
        for d in Direction.all where d != opp {
            if let v = beadTypeId(at: q.neighbor(d), readFromCacheWithoutChecking: wtcwc),
               let r = rule,
               r.attraction(between: typeId, and: v) {
                setBond(at: q, to: d, value: bond, writeToCacheWithoutChecking: wtcwc)
                bondCount += 1
            } else {
                setBond(at: q, to: d, value: nil, writeToCacheWithoutChecking: wtcwc)
            }
        }
        
        setBond(at: p, to: to, value: isTest ? (isSeed ? .testSeed : .testPrimary) : (isSeed ? .seed : .primary), writeToCacheWithoutChecking: wtcwc)
        setOutEdge(at: p, to: to, writeToCacheWithoutChecking: wtcwc)
        
        return bondCount

    }
    
    func addBead(typeId: BeadTypeId, at p: IntPoint, bond: Bond,
                 writeToCacheWithoutChecking wtcwc: Bool) -> Int {
        var bondCount = 0
        setBeadTypeId(at: p, value: typeId, writeToCacheWithoutChecking: wtcwc)
        for d in Direction.all {
            if let v = beadTypeId(at: p.neighbor(d), readFromCacheWithoutChecking: wtcwc),
               let r = rule,
               r.attraction(between: typeId, and: v) {
                setBond(at: p, to: d, value: bond, writeToCacheWithoutChecking: wtcwc)
                bondCount += 1
            } else {
                setBond(at: p, to: d, value: nil, writeToCacheWithoutChecking: wtcwc)
            }
        }
        return bondCount
    }
    
    func removeBead(at p: IntPoint,
                    writeToCacheWithoutChecking wtcwc: Bool) {
        setBeadTypeId(at: p, value: nil, writeToCacheWithoutChecking: wtcwc)
        for d in Direction.all {
            setBond(at: p, to: d, value: nil, writeToCacheWithoutChecking: wtcwc)
        }
        setOutEdge(at: p, to: nil, writeToCacheWithoutChecking: wtcwc)
    }
    
    func bondFor(test: Bool, seed: Bool) -> Bond {
        return test ? .testBond : .bond
    }
    
    func primaryBondFor(test: Bool, seed: Bool) -> Bond {
        return test ? (seed ? .testSeed : .testPrimary) : (seed ? .seed : .primary)
    }
    
    func draw(path: Path, beadTypeStream: (Int) -> BeadTypeId, isTesting: Bool, isSeed: Bool, from p: IntPoint, in range: Range<Int>?,
              writeToCacheWithoutChecking wtcwc: Bool) -> (nbBonds: Int, endPoint: IntPoint) {
        var q = p
        var bq = beadTypeId(at: q, readFromCacheWithoutChecking: wtcwc)
        var idx = beadIndex(at: p, readFromCacheWithoutChecking: wtcwc) ?? 0
        var bondCount = 0
        let bond : Bond = bondFor(test: isTesting, seed: isSeed)
        let primary : Bond = primaryBondFor(test: isTesting, seed: isSeed)

        let range = range ?? 0..<path.count
        for i in range {
            let d = path[i]
            let v = q.neighbor(d)
            let bi = beadTypeStream(i)
            bondCount += addBead(typeId: bi, at: v, bond: bond, writeToCacheWithoutChecking: wtcwc)
            if let b = bq {
                if let r = rule, r.attraction(between: b, and: bi) {
                    bondCount -= 1
                }
                setBond(at: q, to: d, value: primary, writeToCacheWithoutChecking: wtcwc)
                setOutEdge(at: q, to: d, writeToCacheWithoutChecking: wtcwc)
            }
            idx += 1
            setBeadIndex(at: v, value: idx, writeToCacheWithoutChecking: wtcwc)
            q = v
            bq = bi
        }
        return (bondCount, q)
    }
    
    func draw(conformation: Conformation, isTesting: Bool, isSeed: Bool, from p: IntPoint,
              writeToCacheWithoutChecking wtcwc: Bool) -> (nbBonds: Int, endPoint: IntPoint) {
        var q = p
        var bq = beadTypeId(at: q, readFromCacheWithoutChecking: wtcwc)
        var idx = beadIndex(at: p, readFromCacheWithoutChecking: wtcwc) ?? 0
        var bondCount = 0
        let bond : Bond = bondFor(test: isTesting, seed: isSeed)
        let primary : Bond = primaryBondFor(test: isTesting, seed: isSeed)

        for (path, beads) in conformation {
            let expanded_path = path.expand()
            guard expanded_path.count == beads.count else {
                print("In conformation \(conformation), the lengths of path \(path) and beads \(beads) do not match")
                return (bondCount, q)
            }
            for i in 0..<expanded_path.count {
                let d = expanded_path[i]
                let v = q.neighbor(d)
                let bi = beads[i]
            bondCount += addBead(typeId: bi, at: v, bond: bond, writeToCacheWithoutChecking: wtcwc)
                if let b = bq {
                    if let r = rule, r.attraction(between: b, and: bi) {
                        bondCount -= 1
                    }
                    setBond(at: q, to: d, value: primary, writeToCacheWithoutChecking: wtcwc)
                    setOutEdge(at: q, to: d, writeToCacheWithoutChecking: wtcwc)
                }
                idx += 1
                setBeadIndex(at: v, value: idx, writeToCacheWithoutChecking: wtcwc)
                q = v
                bq = bi
            }
        }
        return (bondCount, q)
    }
    
    func erase(path: Path, from p: IntPoint, shortenBy start: Int,
               writeToCacheWithoutChecking wtcwc: Bool) {
        var q = p
        for i in start..<path.count-start {
            q = q.neighbor(path[i])
            removeBead(at: q, writeToCacheWithoutChecking: wtcwc)
        }
    }
    
    func beadTypeStream() -> [BeadTypeId] {
        guard pathStartingPoints.count > 0 else { return [] }
        var s : [BeadTypeId] = []
        synchronize {
            var p = pathStartingPoints[0]
            s = [beadTypeId(at:p)!]
            while let d = outEdge(at: p) {
                p = p.neighbor(d)
                s.append(beadTypeId(at: p)!)
            }
        }
        return s
    }
    
    func beadPath() -> Path {
        guard pathStartingPoints.count > 0 else { return [] }
        var p = pathStartingPoints[0]
        var path : Path = []
        while let d = outEdge(at: p) {
            path.append(d)
            p = p.neighbor(d)
        }
        return path
    }

    func pathBounds()-> CGRect {
        guard pathStartingPoints.count > 0 else { return CGRect.zero }
        var p = pathStartingPoints[0]
        let q = p.point
        var xmin = q.x
        var xmax = q.x
        var ymin = q.y
        var ymax = q.y
        
        while let d = outEdge(at: p) {
            p = p.neighbor(d)
            let q = p.point
            if q.x < xmin { xmin = q.x }
            if q.x > xmax { xmax = q.x }
            if q.y < ymin { ymin = q.y }
            if q.y > ymax { ymax = q.y }
        }
        
        return CGRect(x: xmin, y: ymin, width: xmax-xmin, height: ymax-ymin)
    }
    
    func expand(rect: CGRect, byX x: Int, y: Int) -> CGRect {
        let p = IntPoint(closestTo: CGPoint(x: rect.minX, y: rect.minY))
        let q = IntPoint(closestTo: CGPoint(x: rect.maxX, y: rect.maxY))
        
        let pp = IntPoint(x: p.x - x, y: p.y - y).point
        let qq = IntPoint(x: q.x + x, y: q.y + y).point
        
        return CGRect(x: pp.x, y: pp.y, width: qq.x - pp.x, height: qq.y - pp.y)
    }
        
    static func boundingRect(for path: Path) -> CGRect {
        var (minX, maxX, minY, maxY) : (CGFloat,CGFloat,CGFloat,CGFloat) = (0, 0, 0, 0)
        var q = IntPoint.zero
        for d in path {
            q = q.neighbor(d)
            let p = q.point
            if p.x < minX { minX = p.x }
            if p.x > maxX { maxX = p.x }
            if p.y < minY { minY = p.y }
            if p.y > maxY { maxY = p.y }
        }
        return CGRect(x: minX, y: minY, width: maxX-minX, height: maxY-minY)
    }
    
    func occupiedNeighborDirection(from p: IntPoint,
                                   readFromCacheWithoutChecking rfcwc: Bool) -> Direction {
        for d in Direction.all {
            let v = p.neighbor(d)
            if  isOccupied(at: v, readFromCacheWithoutChecking: rfcwc) {
                return d
            }
        }
        return Direction(rawValue: 0)!
    }
    
    // MARK: unaccessible beads flushing
    
    func get_inaccessible_beads_outside_cache() -> [(IntPoint,IntPoint)] {
        var inaccessible_beads = [(IntPoint,IntPoint)]()
        let db = dict_bitsize
        for (p, dict) in dict_of_dict {
            if let c = last_flush_dict_count[p], c == dict.keys.count {
                // no changes in dict, skip
                continue
            }
            print("flushing dictionary \(p)...")
            let before = inaccessible_beads.count
            let _p = IntPoint(x: p.x<<db, y: p.y<<db)
            for _q in dict.keys {
                let q = IntPoint(x: _q.x|_p.x, y: _q.y|_p.y)
                if isOccupied(at: q) && !cache_contains(q){
                    var flush_q = true
                    for d in Direction.all {
                        if !isOccupied(at: q.neighbor(d)) {
                            flush_q = false
                            break
                        }
                    }
                    if flush_q {
                        inaccessible_beads.append((p,_q))
                    }
                }
            }
            print("... \(inaccessible_beads.count-before) points to be flushed in dictionary \(p)")
        }
        return inaccessible_beads
    }
    
    func update_last_flush_dict_count() {
        for (p, dict) in dict_of_dict {
            last_flush_dict_count[p] = dict.count
        }
    }
    
    func flush_inaccessible_beads_in_all_dictionnaries() {
        synchronizeIfNotAlreadySynchronized {
            push_committed_points()
            let beads_to_delete = get_inaccessible_beads_outside_cache()
            for (p,_q) in beads_to_delete {
                dict_of_dict[p]![_q] = nil
            }
            update_last_flush_dict_count()
            fill_cache_cells()
        }
    }
}
